import numpy as np

from glycresoft.structure import FragmentMatchMap

from glycresoft.tandem.glycopeptide.core_search import approximate_internal_size_of_glycan

from .base import (
    GlycopeptideSpectrumMatcherBase, ChemicalShift, EXDFragmentationStrategy,
    HCDFragmentationStrategy, IonSeries)

from .glycan_signature_ions import GlycanCompositionSignatureMatcher


class SimpleCoverageScorer(GlycopeptideSpectrumMatcherBase):
    backbone_weight = 0.5
    glycosylated_weight = 0.5
    stub_weight = 0.2

    def __init__(self, scan, sequence, mass_shift=None):
        super(SimpleCoverageScorer, self).__init__(scan, sequence, mass_shift)
        self._score = None
        self.solution_map = FragmentMatchMap()
        self.glycosylated_n_term_ion_count = 0
        self.glycosylated_c_term_ion_count = 0

    @property
    def glycosylated_b_ion_count(self):
        return self.glycosylated_n_term_ion_count

    @glycosylated_b_ion_count.setter
    def glycosylated_b_ion_count(self, value):
        self.glycosylated_n_term_ion_count = value

    @property
    def glycosylated_y_ion_count(self):
        return self.glycosylated_c_term_ion_count

    @glycosylated_y_ion_count.setter
    def glycosylated_y_ion_count(self, value):
        self.glycosylated_c_term_ion_count = value

    def _match_backbone_series(self, series, error_tolerance=2e-5, masked_peaks=None, strategy=None,
                               include_neutral_losses=False):
        if strategy is None:
            strategy = HCDFragmentationStrategy
        # Assumes that fragmentation proceeds from the start of the ladder (series position 1)
        # which means that if the last fragment could be glycosylated then the next one will be
        # but if the last fragment wasn't the next one might be.
        previous_position_glycosylated = False
        for frags in self.get_fragments(series, strategy=strategy, include_neutral_losses=include_neutral_losses):
            glycosylated_position = previous_position_glycosylated
            for frag in frags:
                if not glycosylated_position:
                    glycosylated_position |= frag.is_glycosylated
                for peak in self.spectrum.all_peaks_for(frag.mass, error_tolerance):
                    if peak.index.neutral_mass in masked_peaks:
                        continue
                    self.solution_map.add(peak, frag)
            if glycosylated_position:
                if series.direction > 0:
                    self.glycosylated_n_term_ion_count += 1
                else:
                    self.glycosylated_c_term_ion_count += 1
            previous_position_glycosylated = glycosylated_position

    def _compute_coverage_vectors(self):
        n_term_ions = np.zeros(len(self.target))
        c_term_ions = np.zeros(len(self.target))
        stub_count = 0
        glycosylated_n_term_ions = set()
        glycosylated_c_term_ions = set()

        for frag in self.solution_map.fragments():
            series = frag.get_series()
            if series in (IonSeries.b, IonSeries.c):
                n_term_ions[frag.position] = 1
                if frag.is_glycosylated:
                    glycosylated_n_term_ions.add((series, frag.position))
            elif series in (IonSeries.y, IonSeries.z):
                c_term_ions[frag.position] = 1
                if frag.is_glycosylated:
                    glycosylated_c_term_ions.add((series, frag.position))
            elif series == IonSeries.stub_glycopeptide:
                stub_count += 1

        n_term_ions[0] = 0
        return n_term_ions, c_term_ions, stub_count, len(glycosylated_n_term_ions), len(glycosylated_c_term_ions)

    def _compute_glycosylated_coverage(self, glycosylated_n_term_ions, glycosylated_c_term_ions):
        ladders = 0.
        numer = 0.0
        denom = 0.0
        if self.glycosylated_n_term_ion_count > 0:
            numer += glycosylated_n_term_ions
            denom += self.glycosylated_n_term_ion_count
            ladders += 1.
        if self.glycosylated_c_term_ion_count > 0:
            numer += glycosylated_c_term_ions
            denom += self.glycosylated_c_term_ion_count
            ladders += 1.
        if denom == 0.0:
            return 0.0
        return numer / denom

    def _get_internal_size(self, glycan_composition):
        return approximate_internal_size_of_glycan(glycan_composition)

    def compute_coverage(self):
        (n_term_ions, c_term_ions, stub_count,
         glycosylated_n_term_ions,
         glycosylated_c_term_ions) = self._compute_coverage_vectors()

        mean_coverage = np.mean(np.log2(n_term_ions + c_term_ions[::-1] + 1) / np.log2(3))

        glycosylated_coverage = self._compute_glycosylated_coverage(
            glycosylated_n_term_ions,
            glycosylated_c_term_ions)

        stub_fraction = min(stub_count, 3) / 3.

        return mean_coverage, glycosylated_coverage, stub_fraction

    @classmethod
    def get_params(self, backbone_weight=None, glycosylated_weight=None, stub_weight=None, **kwargs):
        if backbone_weight is None:
            backbone_weight = self.backbone_weight
        if glycosylated_weight is None:
            glycosylated_weight = self.glycosylated_weight
        if stub_weight is None:
            stub_weight = self.stub_weight
        return backbone_weight, glycosylated_weight, stub_weight, kwargs

    def calculate_score(self, backbone_weight=None, glycosylated_weight=None, stub_weight=None, **kwargs):
        if backbone_weight is None:
            backbone_weight = self.backbone_weight
        if glycosylated_weight is None:
            glycosylated_weight = self.glycosylated_weight
        if stub_weight is None:
            stub_weight = self.stub_weight
        score = self._coverage_score(backbone_weight, glycosylated_weight, stub_weight)
        self._score = score
        return score

    def _coverage_score(self, backbone_weight=None, glycosylated_weight=None, stub_weight=None):
        if backbone_weight is None:
            backbone_weight = self.backbone_weight
        if glycosylated_weight is None:
            glycosylated_weight = self.glycosylated_weight
        if stub_weight is None:
            stub_weight = self.stub_weight
        mean_coverage, glycosylated_coverage, stub_fraction = self.compute_coverage()
        score = (((mean_coverage * backbone_weight) + (glycosylated_coverage * glycosylated_weight)) * (
            1 - stub_weight)) + (stub_fraction * stub_weight)
        return score


class SignatureAwareCoverageScorer(SimpleCoverageScorer, GlycanCompositionSignatureMatcher):
    def __init__(self, scan, sequence, mass_shift=None, *args, **kwargs):
        super(SignatureAwareCoverageScorer, self).__init__(scan, sequence, mass_shift, *args, **kwargs)

    def match(self, error_tolerance=2e-5, *args, **kwargs):
        GlycanCompositionSignatureMatcher.match(self, error_tolerance=error_tolerance, **kwargs)
        masked_peaks = set()
        include_neutral_losses = kwargs.get("include_neutral_losses", False)
        extended_glycan_search = kwargs.get("extended_glycan_search", False)

        if self.mass_shift.tandem_mass != 0:
            chemical_shift = ChemicalShift(
                self.mass_shift.name, self.mass_shift.tandem_composition)
        else:
            chemical_shift = None

        is_hcd = self.is_hcd()
        is_exd = self.is_exd()
        if not is_hcd and not is_exd:
            is_hcd = True
        # handle glycan fragments from collisional dissociation
        if is_hcd:
            self._match_oxonium_ions(error_tolerance, masked_peaks=masked_peaks)
            self._match_stub_glycopeptides(error_tolerance, masked_peaks=masked_peaks,
                                           chemical_shift=chemical_shift,
                                           extended_glycan_search=extended_glycan_search)
        # handle N-term
        if is_hcd and not is_exd:
            self._match_backbone_series(
                IonSeries.b, error_tolerance, masked_peaks, HCDFragmentationStrategy, include_neutral_losses)
        elif is_exd:
            self._match_backbone_series(
                IonSeries.b, error_tolerance, masked_peaks, EXDFragmentationStrategy, include_neutral_losses)
            self._match_backbone_series(
                IonSeries.c, error_tolerance, masked_peaks, EXDFragmentationStrategy, include_neutral_losses)

        # handle C-term
        if is_hcd and not is_exd:
            self._match_backbone_series(
                IonSeries.y, error_tolerance, masked_peaks, HCDFragmentationStrategy, include_neutral_losses)
        elif is_exd:
            self._match_backbone_series(
                IonSeries.y, error_tolerance, masked_peaks, EXDFragmentationStrategy, include_neutral_losses)
            self._match_backbone_series(
                IonSeries.z, error_tolerance, masked_peaks, EXDFragmentationStrategy, include_neutral_losses)
        return self


try:
    from glycresoft._c.tandem.tandem_scoring_helpers import (
        _compute_coverage_vectors, SimpleCoverageScorer_match_backbone_series)
    SimpleCoverageScorer._compute_coverage_vectors = _compute_coverage_vectors
    SimpleCoverageScorer._match_backbone_series = SimpleCoverageScorer_match_backbone_series
except ImportError:
    pass



from typing import Callable, List, Optional, Tuple

from ms_deisotope.peak_set import DeconvolutedPeak
from ms_deisotope.peak_dependency_network.intervals import Interval, IntervalTreeNode

from glycresoft.task import TaskBase

from .chromatogram import Chromatogram


class ChromatogramForest(TaskBase):
    """An an algorithm for aggregating chromatograms from peaks of close mass
    weighted by intensity.

    This algorithm assumes that mass accuracy is correlated with intensity, so
    the most intense peaks should most accurately reflect their true neutral mass.
    The expected input is a list of (scan id, peak) pairs. This list is sorted by
    descending peak intensity. For each pair, using binary search, locate the nearest
    existing chromatogram in :attr:`chromatograms`. If the nearest chromatogram is within
    :attr:`error_tolerance` ppm of the peak's neutral mass, add this peak to that
    chromatogram, otherwise create a new chromatogram containing this peak and insert
    it into :attr:`chromatograms` while preserving the overall sortedness. This algorithm
    is carried out by :meth:`aggregate_unmatched_peaks`

    This process may produce chromatograms with large gaps in them, which
    may or may not be acceptable. To break gapped chromatograms into separate
    entities, the :class:`ChromatogramFilter` type has a method :meth:`split_sparse`.

    Attributes
    ----------
    chromatograms : list of Chromatogram
        A list of growing Chromatogram objects, ordered by neutral mass
    count : int
        The number of peaks accumulated
    error_tolerance : float
        The mass error tolerance between peaks and possible chromatograms (in ppm)
    scan_id_to_rt : callable
        A callable object to convert scan ids to retention time.
    count : int
        The number of peaks handled
    """
    chromatograms: List[Chromatogram]
    error_tolerance: float
    scan_id_to_rt: Callable[[str], float]
    count: int

    def __init__(self, chromatograms=None, error_tolerance=1e-5, scan_id_to_rt=lambda x: x):
        if chromatograms is None:
            chromatograms = []
        self.chromatograms = sorted(chromatograms, key=lambda x: x.neutral_mass)
        self.error_tolerance = error_tolerance
        self.scan_id_to_rt = scan_id_to_rt
        self.count = 0

    def __len__(self):
        return len(self.chromatograms)

    def __iter__(self):
        return iter(self.chromatograms)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.chromatograms[i]
        else:
            return [self.chromatograms[j] for j in i]

    def find_insertion_point(self, peak: DeconvolutedPeak) -> Tuple[List[int], bool]:
        index, matched = binary_search_with_flag(
            self.chromatograms, peak.neutral_mass, self.error_tolerance)
        return index, matched

    def find_minimizing_index(self, peak: DeconvolutedPeak, indices: List[int]) -> int:
        best_index = None
        best_error = float('inf')
        for index_case in indices:
            chroma = self[index_case]
            err = abs(chroma.neutral_mass - peak.neutral_mass) / peak.neutral_mass
            if err < best_error:
                best_index = index_case
                best_error = err
        return best_index

    def handle_peak(self, scan_id: str, peak: DeconvolutedPeak):
        if len(self) == 0:
            index = [0]
            matched = False
        else:
            index, matched = self.find_insertion_point(peak)
        if matched:
            chroma = self.chromatograms[self.find_minimizing_index(peak, index)]
            most_abundant_member = chroma.most_abundant_member
            chroma.insert(scan_id, peak, self.scan_id_to_rt(scan_id))
            if peak.intensity < most_abundant_member:
                chroma.retain_most_abundant_member()
        else:
            chroma = Chromatogram(None)
            chroma.created_at = "forest"
            chroma.insert(scan_id, peak, self.scan_id_to_rt(scan_id))
            self.insert_chromatogram(chroma, index)
        self.count += 1

    def insert_chromatogram(self, chromatogram: Chromatogram, index: Tuple[int, bool]):
        # TODO: Review this index arithmetic, the output isn't sorted.
        index = index[0] # index is (index, matched) from binary_search_with_flag
        if index != 0:
            self.chromatograms.insert(index + 1, chromatogram)
        else:
            if len(self) == 0:
                new_index = index
            else:
                x = self.chromatograms[index]
                if x.neutral_mass < chromatogram.neutral_mass:
                    new_index = index + 1
                else:
                    new_index = index
            self.chromatograms.insert(new_index, chromatogram)

    def aggregate_unmatched_peaks(self, *args, **kwargs):
        import warnings
        warnings.warn("Instead of calling aggregate_unmatched_peaks, call aggregate_peaks", stacklevel=2)
        self.aggregate_peaks(*args, **kwargs)

    def aggregate_peaks(self, scan_id_peaks_list, minimum_mass=300, minimum_intensity=1000.):
        unmatched = sorted(scan_id_peaks_list, key=lambda x: x[1].intensity, reverse=True)
        for scan_id, peak in unmatched:
            if peak.neutral_mass < minimum_mass or peak.intensity < minimum_intensity:
                continue
            self.handle_peak(scan_id, peak)


class ChromatogramMerger(TaskBase):
    def __init__(self, chromatograms=None, error_tolerance=1e-5):
        if chromatograms is None:
            chromatograms = []
        self.chromatograms = sorted(chromatograms, key=lambda x: x.neutral_mass)
        self.error_tolerance = error_tolerance
        self.count = 0
        self.verbose = False

    def __len__(self):
        return len(self.chromatograms)

    def __iter__(self):
        return iter(self.chromatograms)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.chromatograms[i]
        else:
            return [self.chromatograms[j] for j in i]

    def find_candidates(self, new_chromatogram: Chromatogram) -> Tuple[List[int], bool]:
        index, matched = binary_search_with_flag(
            self.chromatograms, new_chromatogram.neutral_mass, self.error_tolerance)
        return index, matched

    def merge_overlaps(self, new_chromatogram: Chromatogram, chromatogram_range: List[Chromatogram]) -> bool:
        has_merged = False
        query_mass = new_chromatogram.neutral_mass
        for chroma in chromatogram_range:
            cond = (chroma.overlaps_in_time(new_chromatogram) and abs(
                    (chroma.neutral_mass - query_mass) / query_mass) < self.error_tolerance and
                    not chroma.common_nodes(new_chromatogram))
            if cond:
                chroma.merge(new_chromatogram)
                has_merged = True
                break
        return has_merged

    def find_insertion_point(self, new_chromatogram: Chromatogram) -> Optional[int]:
        return binary_search_exact(
            self.chromatograms, new_chromatogram.neutral_mass)

    def handle_new_chromatogram(self, new_chromatogram: Chromatogram):
        if len(self) == 0:
            index = [0]
            matched = False
        else:
            index, matched = self.find_candidates(new_chromatogram)
        if matched:

            chroma = self[index]
            has_merged = self.merge_overlaps(new_chromatogram, chroma)
            if not has_merged:
                insertion_point = self.find_insertion_point(new_chromatogram)
                self.insert_chromatogram(new_chromatogram, [insertion_point])
        else:
            self.insert_chromatogram(new_chromatogram, index)
        self.count += 1

    def insert_chromatogram(self, chromatogram: Chromatogram, index: List[int]):
        if index[0] != 0:
            self.chromatograms.insert(index[0] + 1, chromatogram)
        else:
            if len(self) == 0:
                new_index = index[0]
            else:
                x = self.chromatograms[index[0]]
                if x.neutral_mass < chromatogram.neutral_mass:
                    new_index = index[0] + 1
                else:
                    new_index = index[0]
            self.chromatograms.insert(new_index, chromatogram)

    def aggregate_chromatograms(self, chromatograms):
        unmatched = sorted(chromatograms, key=lambda x: x.total_signal, reverse=True)
        for chroma in unmatched:
            self.handle_new_chromatogram(chroma)


def flatten_tree(tree):
    output_queue = []
    input_queue = [tree]
    while input_queue:
        next_node = input_queue.pop()
        output_queue.append(next_node)

        next_right = next_node.right
        if next_right is not None:
            input_queue.append(next_right)

        next_left = next_node.left
        if next_left is not None:
            input_queue.append(next_left)
    return output_queue[::-1]


def layered_traversal(nodes):
    return sorted(nodes, key=lambda x: (x.level, x.center), reverse=True)


class ChromatogramOverlapSmoother(object):
    def __init__(self, chromatograms, error_tolerance=1e-5):
        self.retention_interval_tree = build_rt_interval_tree(chromatograms)
        self.error_tolerance = error_tolerance
        self.solution_map = {None: []}
        self.chromatograms = self.smooth()

    def __iter__(self):
        return iter(self.chromatograms)

    def __getitem__(self, i):
        return self.chromatograms[i]

    def __len__(self):
        return len(self.chromatograms)

    def aggregate_interval(self, tree):
        chromatograms = [interval[0] for interval in tree.contained]
        chromatograms.extend(self.solution_map[tree.left])
        chromatograms.extend(self.solution_map[tree.right])
        merger = ChromatogramMerger(error_tolerance=self.error_tolerance)
        merger.aggregate_chromatograms(chromatograms)
        self.solution_map[tree] = list(merger)
        return merger

    def smooth(self):
        nodes = layered_traversal(flatten_tree(self.retention_interval_tree))
        for node in nodes:
            self.aggregate_interval(node)
        final = self.solution_map[self.retention_interval_tree]
        result = ChromatogramMerger()
        result.aggregate_chromatograms(final)
        return list(result)


def binary_search_with_flag(array, mass, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - mass) / mass
        if abs(err) <= error_tolerance:
            i = mid - 1
            # Begin Sweep forward
            while i > 0:
                x = array[i]
                err = (x.neutral_mass - mass) / mass
                if abs(err) <= error_tolerance:
                    i -= 1
                    continue
                else:
                    break
            low_end = i
            i = mid + 1

            # Begin Sweep backward
            while i < n:
                x = array[i]
                err = (x.neutral_mass - mass) / mass
                if abs(err) <= error_tolerance:
                    i += 1
                    continue
                else:
                    break
            high_end = i
            return list(range(low_end, high_end)), True
        elif (hi - lo) == 1:
            return [mid], False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0, False


def binary_search_exact(array, mass):
    lo = 0
    hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - mass)
        if err == 0:
            return mid
        elif (hi - lo) == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid


def smooth_overlaps(chromatogram_list, error_tolerance=1e-5):
    chromatogram_list = sorted(chromatogram_list, key=lambda x: x.neutral_mass)
    out = []
    last = chromatogram_list[0]
    i = 1
    while i < len(chromatogram_list):
        current = chromatogram_list[i]
        mass_error = abs((last.neutral_mass - current.neutral_mass) / current.neutral_mass)
        if mass_error <= error_tolerance:
            if last.overlaps_in_time(current):
                last = last.merge(current)
                last.created_at = "smooth_overlaps"
            else:
                out.append(last)
                last = current
        else:
            out.append(last)
            last = current
        i += 1
    out.append(last)
    return out


class ChromatogramRetentionTimeInterval(Interval):
    def __init__(self, chromatogram):
        super(ChromatogramRetentionTimeInterval, self).__init__(
            chromatogram.start_time, chromatogram.end_time, [chromatogram])
        self.neutral_mass = chromatogram.neutral_mass
        self.start_time = self.start
        self.end_time = self.end
        self.data['neutral_mass'] = self.neutral_mass


def build_rt_interval_tree(chromatogram_list, interval_tree_type=IntervalTreeNode):
    intervals = list(map(ChromatogramRetentionTimeInterval, chromatogram_list))
    interval_tree = interval_tree_type.build(intervals)
    return interval_tree



from glycresoft.scan_cache import NullScanStorageHandler


class ScanSink(object):
    def __init__(self, scan_generator, cache_handler_type=NullScanStorageHandler):
        self.scan_generator = scan_generator
        self.scan_store = None
        self._scan_store_type = cache_handler_type

    @property
    def scan_source(self):
        try:
            return self.scan_generator.scan_source
        except AttributeError:
            return None

    @property
    def sample_run(self):
        try:
            return self.scan_store.sample_run
        except AttributeError:
            return None

    def configure_cache(self, storage_path=None, name=None, source=None):
        self.scan_store = self._scan_store_type.configure_storage(
            storage_path, name, source)

    def configure_iteration(self, *args, **kwargs):
        self.scan_generator.configure_iteration(*args, **kwargs)

    def scan_id_to_rt(self, scan_id):
        return self.scan_generator.convert_scan_id_to_retention_time(scan_id)

    def store_scan(self, scan):
        if self.scan_store is not None:
            self.scan_store.accumulate(scan)

    def commit(self):
        if self.scan_store is not None:
            self.scan_store.commit()

    def complete(self):
        if self.scan_store is not None:
            self.scan_store.complete()
        self.scan_generator.close()

    def next_scan(self):
        scan = next(self.scan_generator)
        self.store_scan(scan)
        while scan.ms_level != 1:
            scan = next(self.scan_generator)
            self.store_scan(scan)
        return scan

    def __iter__(self):
        return self

    def __next__(self):
        return self.next_scan()

    def next(self):
        return self.next_scan()



from .base import (DEBUG_MODE, Predicate, default_threshold, always, TargetType, logger,
                   SolutionEntry, build_glycopeptide_key, KeyTargetMassShift, KeyMassShift, parsimony_sort)
from .revision import (MS2RevisionValidator, SignalUtilizationMS2RevisionValidator,
                       SpectrumMatchBackFiller, SpectrumMatchUpdater, RevisionSummary)
from .graph import (RepresenterDeconvolution, MassShiftDeconvolutionGraph,
                    MassShiftDeconvolutionGraphNode, MassShiftDeconvolutionGraphEdge)
from .representer import (RepresenterSelectionStrategy,
                          TotalBestRepresenterStrategy, TotalAboveAverageBestRepresenterStrategy)
from .chromatogram import (TandemAnnotatedChromatogram, SpectrumMatchSolutionCollectionBase,
                           TandemSolutionsWithoutChromatogram, ScanTimeBundle)
from .aggregation import (aggregate_by_assigned_entity,
                          AnnotatedChromatogramAggregator, GraphAnnotatedChromatogramAggregator)
from .mapper import ChromatogramMSMSMapper

__all__ = [
    'DEBUG_MODE', 'Predicate', 'default_threshold', 'always', 'TargetType', 'logger',
    'SolutionEntry', 'build_glycopeptide_key', 'KeyTargetMassShift', 'KeyMassShift', 'parsimony_sort',
    'MS2RevisionValidator', 'SignalUtilizationMS2RevisionValidator', 'SpectrumMatchBackFiller',
    'SpectrumMatchUpdater', 'RepresenterDeconvolution', 'MassShiftDeconvolutionGraph',
    'MassShiftDeconvolutionGraphNode', 'MassShiftDeconvolutionGraphEdge',
    'RepresenterSelectionStrategy', 'TotalBestRepresenterStrategy', 'TotalAboveAverageBestRepresenterStrategy',
    'TandemAnnotatedChromatogram', 'SpectrumMatchSolutionCollectionBase', 'TandemSolutionsWithoutChromatogram',
    'ScanTimeBundle', 'aggregate_by_assigned_entity', 'AnnotatedChromatogramAggregator',
    'GraphAnnotatedChromatogramAggregator', 'ChromatogramMSMSMapper', "RevisionSummary"
]



import os
from uuid import uuid4

import dill as pickle

import click

from ms_deisotope.data_source.metadata.sample import Sample

from glycresoft import serialize
from glycresoft.cli.export import analysis_identifier_arg

from glycresoft.serialize import (
    DatabaseBoundOperation,
    GlycanHypothesis,
    GlycopeptideHypothesis,
    Analysis)

from glycresoft.profiler import (
    MzMLGlycopeptideLCMSMSAnalyzer,
    MzMLComparisonGlycopeptideLCMSMSAnalyzer,
    MultipartGlycopeptideLCMSMSAnalyzer,
    GlycopeptideFDREstimationStrategy,
    MzMLGlycanChromatogramAnalyzer,
    LaplacianRegularizedChromatogramProcessor,
    ProcessedMSFileLoader,
    ChromatogramSummarizer)

from glycresoft.composition_distribution_model import site_model
from glycresoft.scoring.elution_time_grouping import GlycopeptideChromatogramProxy, GlycopeptideElutionTimeModeler
from glycresoft.serialize.analysis import AnalysisTypeEnum
from glycresoft.tandem.glycopeptide.scoring import CoverageWeightedBinomialScorer
from glycresoft.composition_distribution_model import GridPointSolution
from glycresoft.database.composition_network import GraphReader

from glycresoft.models import GeneralScorer
from glycresoft.task import fmt_msg

from .base import (
    cli, HiddenOption, processes_option)

from .validators import (
    validate_analysis_name,
    validate_mass_shift, validate_glycopeptide_tandem_scoring_function,
    glycopeptide_tandem_scoring_functions,
    get_by_name_or_id,
    validate_ms1_feature_name,
    ms1_model_features,
    RelativeMassErrorParam,
    DatabaseConnectionParam,
    SmallSampleFDRCorrectionParam, ChoiceOrURI)


def make_analysis_output_path(prefix):
    output_path_pattern = prefix + "_analysis_%d.db"
    i = 1
    while os.path.exists(output_path_pattern % i) and i < (2 ** 16):
        i += 1
    if i >= (2 ** 16):
        return output_path_pattern % (uuid4().int,)
    else:
        return output_path_pattern % (i,)


@cli.group(short_help='Identify structures in preprocessed data')
def analyze():
    pass


def database_connection_arg(arg_name='database-connection'):
    def database_connection_arg(fn):
        arg = click.argument(
            arg_name,
            type=DatabaseConnectionParam(exists=True),
            doc_help=(
                "A connection URI for a database, or a path on the file system"))
        return arg(fn)
    return database_connection_arg


def hypothesis_identifier_arg(hypothesis_type, arg_name='hypothesis-identifier'):
    def wrapper(fn):
        arg = click.argument(arg_name, doc_help=(
            "The ID number or name of the %s hypothesis to use" % (hypothesis_type,)))
        return arg(fn)
    return wrapper


def hypothesis_identifier_arg_option(hypothesis_type, *args, **kwargs):
    def wrapper(fn):
        arg = click.option(*args, help=(
            "The ID number or name of the %s hypothesis to use" % (hypothesis_type,)), **kwargs)
        return arg(fn)
    return wrapper


def sample_path_arg(fn):
    arg = click.argument(
        "sample-path",
        type=click.Path(exists=True, dir_okay=False, file_okay=True),
        doc_help=(
            "The path to the deconvoluted sample file"))
    return arg(fn)


@analyze.command("search-glycopeptide", short_help='Search preprocessed data for glycopeptide sequences')
@click.pass_context
@database_connection_arg()
@sample_path_arg
@hypothesis_identifier_arg("glycopeptide")
@click.option("-m", "--mass-error-tolerance", type=RelativeMassErrorParam(), default=1e-5,
              help="Mass accuracy constraint, in parts-per-million error, for matching MS^1 ions.")
@click.option("-mn", "--msn-mass-error-tolerance", type=RelativeMassErrorParam(), default=2e-5,
              help="Mass accuracy constraint, in parts-per-million error, for matching MS^n ions.")
@click.option("-g", "--grouping-error-tolerance", type=RelativeMassErrorParam(), default=1.5e-5,
              help="Mass accuracy constraint, in parts-per-million error, for grouping chromatograms.")
@click.option("-n", "--analysis-name", default=None, help='Name for analysis to be performed.')
@click.option("-q", "--psm-fdr-threshold", default=0.05, type=float,
              help='Minimum FDR Threshold to use for filtering GPSMs when selecting identified glycopeptides')
@click.option("-s", "--tandem-scoring-model", default='coverage_weighted_binomial',
              type=ChoiceOrURI(glycopeptide_tandem_scoring_functions.keys()),
              help="Select a scoring function to use for evaluating glycopeptide-spectrum matches")
@click.option("-x", "--oxonium-threshold", default=0.05, type=float,
              help=('Minimum HexNAc-derived oxonium ion abundance '
                    'ratio to filter MS/MS scans. Defaults to 0.05.'))
@click.option("-a", "--adduct", 'mass_shifts', multiple=True, nargs=2,
              help=("Adducts to consider. Specify name or formula, and a"
                    " multiplicity."))
@click.option("-f", "--use-peptide-mass-filter", is_flag=True, help=(
    "Filter putative spectrum matches by estimating the peptide backbone mass "
    "from the precursor mass and stub glycopeptide signature ions"))
@processes_option
@click.option("--export", type=click.Choice(['csv', 'html', 'psm-csv']), multiple=True,
              help="export command to after search is complete")
@click.option("-o", "--output-path", default=None, type=click.Path(writable=True),
              help=("Path to write resulting analysis to."), required=True)
@click.option("-w", "--workload-size", default=500, type=int, help="Number of spectra to process at once")
@click.option("--save-intermediate-results", default=None, type=click.Path(), required=False,
              help='Save intermediate spectrum matches to a file')
@click.option("--maximum-mass", default=float('inf'), type=float)
@click.option("-D", "--decoy-database-connection", default=None, help=(
    "Provide an alternative hypothesis to draw decoy glycopeptides from instead of the simpler reversed-peptide "
    "decoy. This is especially necessary when the stub peptide+Y ions account for a large fraction of MS2 signal."))
@click.option("-G", "--permute-decoy-glycan-fragments", is_flag=True, default=False, help=(
    "Whether or not to permute decoy glycopeptides' peptide+Y ions. The intact mass, peptide, "
    "and peptide+Y1 ions are unchanged."))
@click.option("--fdr-correction", default='auto', type=SmallSampleFDRCorrectionParam(),
              help=("Whether to attempt to correct for small sample size for target-decoy analysis."),
              cls=HiddenOption)
@click.option("--isotope-probing-range", type=int, default=3, help=(
    "The maximum number of isotopic peak errors to allow when searching for untrusted precursor masses"))
@click.option("-R", "--rare-signatures", is_flag=True, default=False,
              help="Look for rare signature ions when scoring glycan oxonium signature")
@click.option("--retention-time-modeling/--no-retention-time-modeling", is_flag=True, default=True,
              help=("Whether or not to model relative retention time to correct for common glycan"
                    " composition errors."))
def search_glycopeptide(context, database_connection, sample_path, hypothesis_identifier,
                        analysis_name, output_path=None, grouping_error_tolerance=1.5e-5, mass_error_tolerance=1e-5,
                        msn_mass_error_tolerance=2e-5, psm_fdr_threshold=0.05, peak_shape_scoring_model=None,
                        tandem_scoring_model=None, oxonium_threshold=0.15, save_intermediate_results=None,
                        processes=4, workload_size=500, mass_shifts=None, export=None,
                        use_peptide_mass_filter=False, maximum_mass=float('inf'),
                        decoy_database_connection=None, fdr_correction='auto',
                        isotope_probing_range=3,
                        permute_decoy_glycan_fragments=False,
                        rare_signatures=False,
                        retention_time_modeling=True):
    """Identify glycopeptide sequences from processed LC-MS/MS data. This algorithm requires a fully materialized
    cross-product database (the default), and uses a reverse-peptide decoy by default, evaluated on the total score.

    For a search algorithm that applies separate FDR control on the peptide and the glycan, see
    :ref:`search-glycopeptide-multipart`
    """
    if tandem_scoring_model is None:
        tandem_scoring_model = CoverageWeightedBinomialScorer
    if os.path.exists(output_path):
        click.secho("Output path '%s' exists, removing..." % output_path, fg='yellow')
        os.remove(output_path)
    database_connection = DatabaseBoundOperation(database_connection)
    ms_data = ProcessedMSFileLoader(sample_path, use_index=False)
    try:
        sample_run = ms_data.sample_run
    except StopIteration:
        sample_run = Sample(name=os.path.basename(sample_path).split('.')[0], id=str(uuid4()))

    try:
        hypothesis = get_by_name_or_id(
            database_connection, GlycopeptideHypothesis, hypothesis_identifier)
    except Exception:
        click.secho("Could not locate a Glycopeptide Hypothesis with identifier %r" %
                    hypothesis_identifier, fg='yellow')
        raise click.Abort()

    tandem_scoring_model, evaluation_kwargs = validate_glycopeptide_tandem_scoring_function(
        context, tandem_scoring_model)

    mass_shifts = [validate_mass_shift(mass_shift, multiplicity)
                   for mass_shift, multiplicity in mass_shifts]
    expanded = []
    expanded = MzMLGlycanChromatogramAnalyzer.expand_mass_shifts(
        dict(mass_shifts), crossproduct=False)
    mass_shifts = expanded

    if analysis_name is None:
        analysis_name = "%s @ %s" % (sample_run.name, hypothesis.name)

    analysis_name = validate_analysis_name(
        context, database_connection.session, analysis_name)

    click.secho("Preparing analysis of %s by %s" % (
        sample_run.name, hypothesis.name), fg='cyan')

    if decoy_database_connection is None:
        analyzer = MzMLGlycopeptideLCMSMSAnalyzer(
            database_connection._original_connection,
            sample_path=sample_path,
            hypothesis_id=hypothesis.id,
            analysis_name=analysis_name,
            output_path=output_path,
            grouping_error_tolerance=grouping_error_tolerance,
            mass_error_tolerance=mass_error_tolerance,
            msn_mass_error_tolerance=msn_mass_error_tolerance,
            psm_fdr_threshold=psm_fdr_threshold,
            peak_shape_scoring_model=peak_shape_scoring_model,
            tandem_scoring_model=tandem_scoring_model,
            oxonium_threshold=oxonium_threshold,
            n_processes=processes,
            spectrum_batch_size=workload_size,
            mass_shifts=mass_shifts,
            use_peptide_mass_filter=use_peptide_mass_filter,
            maximum_mass=maximum_mass,
            probing_range_for_missing_precursors=isotope_probing_range,
            permute_decoy_glycans=permute_decoy_glycan_fragments,
            rare_signatures=rare_signatures,
            model_retention_time=retention_time_modeling,
            evaluation_kwargs=evaluation_kwargs
        )
    else:
        analyzer = MzMLComparisonGlycopeptideLCMSMSAnalyzer(
            database_connection._original_connection,
            decoy_database_connection,
            sample_path=sample_path,
            hypothesis_id=hypothesis.id,
            analysis_name=analysis_name,
            output_path=output_path,
            grouping_error_tolerance=grouping_error_tolerance,
            mass_error_tolerance=mass_error_tolerance,
            msn_mass_error_tolerance=msn_mass_error_tolerance,
            psm_fdr_threshold=psm_fdr_threshold,
            peak_shape_scoring_model=peak_shape_scoring_model,
            tandem_scoring_model=tandem_scoring_model,
            oxonium_threshold=oxonium_threshold,
            n_processes=processes,
            spectrum_batch_size=workload_size,
            mass_shifts=mass_shifts,
            use_peptide_mass_filter=use_peptide_mass_filter,
            maximum_mass=maximum_mass,
            use_decoy_correction_threshold=fdr_correction,
            probing_range_for_missing_precursors=isotope_probing_range,
            permute_decoy_glycans=permute_decoy_glycan_fragments,
            rare_signatures=rare_signatures,
            model_retention_time=retention_time_modeling,
            evaluation_kwargs=evaluation_kwargs
        )
    analyzer.display_header()
    result = analyzer.start()
    gps, unassigned, target_decoy_set = result[:3]
    if save_intermediate_results is not None:
        analyzer.log("Saving Intermediate Results")
        with open(save_intermediate_results, 'wb') as handle:
            pickle.dump((target_decoy_set, gps), handle)
    del gps
    del unassigned
    del target_decoy_set
    if export:
        for export_type in set(export):
            click.echo(fmt_msg("Handling Export: %s" % (export_type,)))
            if export_type == 'csv':
                from glycresoft.cli.export import glycopeptide_identification
                base = os.path.splitext(output_path)[0]
                export_path = "%s-glycopeptides.csv" % (base,)
                context.invoke(
                    glycopeptide_identification,
                    database_connection=output_path,
                    analysis_identifier=analyzer.analysis_id,
                    output_path=export_path)
            elif export_type == 'psm-csv':
                from glycresoft.cli.export import glycopeptide_spectrum_matches
                base = os.path.splitext(output_path)[0]
                export_path = "%s-glycopeptide-spectrum-matches.csv" % (base,)
                context.invoke(
                    glycopeptide_spectrum_matches,
                    database_connection=output_path,
                    analysis_identifier=analyzer.analysis_id,
                    output_path=export_path)
            elif export_type == 'html':
                from glycresoft.cli.export import glycopeptide_identification
                base = os.path.splitext(output_path)[0]
                export_path = "%s-report.html" % (base,)
                context.invoke(
                    glycopeptide_identification,
                    database_connection=output_path,
                    analysis_identifier=analyzer.analysis_id,
                    output_path=export_path,
                    report=True)


@analyze.command("search-glycopeptide-multipart", short_help=(
    'Search preprocessed data for glycopeptide sequences scored for both peptide and glycan components'))
@click.pass_context
@database_connection_arg("database-connection")
@database_connection_arg("decoy-database-connection")
@sample_path_arg
@hypothesis_identifier_arg_option("glycopeptide", '-T', '--target-hypothesis-identifier', default=1)
@hypothesis_identifier_arg_option("glycopeptide", '-D', "--decoy-hypothesis-identifier", default=1)
@click.option("-M", "--memory-database-index", is_flag=True,
              help=("Whether to load the entire peptide database into memory during spectrum mapping. "
                    "Uses more memory but substantially accelerates the process"), default=False)
@click.option("-m", "--mass-error-tolerance", type=RelativeMassErrorParam(), default=1e-5,
              help="Mass accuracy constraint, in parts-per-million error, for matching MS^1 ions.")
@click.option("-mn", "--msn-mass-error-tolerance", type=RelativeMassErrorParam(), default=2e-5,
              help="Mass accuracy constraint, in parts-per-million error, for matching MS^n ions.")
@click.option("-g", "--grouping-error-tolerance", type=RelativeMassErrorParam(), default=1.5e-5,
              help="Mass accuracy constraint, in parts-per-million error, for grouping chromatograms.")
@click.option("-n", "--analysis-name", default=None, help='Name for analysis to be performed.')
@click.option("-q", "--psm-fdr-threshold", default=0.05, type=float,
              help='Minimum FDR Threshold to use for filtering GPSMs when selecting identified glycopeptides')
@click.option("-f", "--fdr-estimation-strategy", type=click.Choice(['joint', 'peptide', 'glycan', 'any']),
              default='joint',
              help=("The FDR estimation strategy to use. The joint estimate uses both peptide and glycan scores, "
                    "peptide uses only peptide scores, glycan uses only glycan scores, and any uses the smallest "
                    "FDR of the joint, peptide, and glycan estiamtes."))
@click.option("-s", "--tandem-scoring-model", default='log_intensity',
              type=ChoiceOrURI(["log_intensity",
                                "simple",
                                "penalized_log_intensty",
                                "log_intensity_v3"]),
              help="Select a scoring function to use for evaluating glycopeptide-spectrum matches")
@click.option("-y", "--glycan-score-threshold", default=1.0, type=float,
              help="The minimum glycan score required to consider a peptide mass")
@click.option("-a", "--adduct", 'mass_shifts', multiple=True, nargs=2,
              help=("Adducts to consider. Specify name or formula, and a"
                    " multiplicity."))
@processes_option
@click.option("--export", type=click.Choice(['csv', 'html', 'psm-csv']), multiple=True,
              help="export command to after search is complete")
@click.option("-o", "--output-path", default=None, type=click.Path(writable=True),
              help=("Path to write resulting analysis to."), required=True)
@click.option("-w", "--workload-size", default=100, type=int, help="Number of spectra to process at once")
@click.option("-R", "--rare-signatures", is_flag=True, default=False,
              help="Look for rare signature ions when scoring glycan oxonium signature")
@click.option("--retention-time-modeling/--no-retention-time-modeling", is_flag=True, default=True,
              help=("Whether or not to model relative retention time to correct for common glycan"
                    " composition errors."))
@click.option("--save-intermediate-results", default=None, type=click.Path(), required=False,
              help='Save intermediate spectrum matches to a file', cls=HiddenOption)
@click.option("--maximum-mass", default=float('inf'), type=float, cls=HiddenOption)
@click.option("--isotope-probing-range", type=int, default=3, help=(
    "The maximum number of isotopic peak errors to allow when searching for untrusted precursor masses"))
@click.option("-S", "--glycoproteome-smoothing-model", type=click.Path(readable=True), help=(
    "Path to a glycoproteome site-specific glycome model"), default=None)
@click.option("-x", "--oxonium-threshold", default=0.05, type=float,
              help=('Minimum HexNAc-derived oxonium ion abundance '
                    'ratio to filter MS/MS scans. Defaults to 0.05.'))
@click.option("-P", "--peptide-masses-per-scan", type=int, default=60,
              help="The maximum number of peptide masses to consider per scan")
def search_glycopeptide_multipart(context, database_connection,
                                  decoy_database_connection,
                                  sample_path,
                                  target_hypothesis_identifier=1,
                                  decoy_hypothesis_identifier=1,
                                  analysis_name=None,
                                  output_path=None,
                                  grouping_error_tolerance=1.5e-5,
                                  mass_error_tolerance=1e-5,
                                  msn_mass_error_tolerance=2e-5,
                                  psm_fdr_threshold=0.05,
                                  peak_shape_scoring_model=None,
                                  tandem_scoring_model=None,
                                  glycan_score_threshold=1.0,
                                  memory_database_index=False,
                                  save_intermediate_results=None,
                                  processes=4,
                                  workload_size=100,
                                  mass_shifts=None,
                                  export=None,
                                  maximum_mass=float('inf'),
                                  isotope_probing_range=3,
                                  fdr_estimation_strategy=None,
                                  glycoproteome_smoothing_model=None,
                                  rare_signatures=False,
                                  retention_time_modeling=True,
                                  oxonium_threshold: float=0.05,
                                  peptide_masses_per_scan: int=60):
    """
    Search preprocessed data for glycopeptide sequences scored for both peptide and glycan components.

    This search strategy requires an explicit decoy database, like one created with the `--reverse`
    flag from `glycopeptide-fa`.
    """
    if fdr_estimation_strategy is None:
        fdr_estimation_strategy = GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture
    else:
        fdr_estimation_strategy = GlycopeptideFDREstimationStrategy[fdr_estimation_strategy]
    if tandem_scoring_model is None:
        tandem_scoring_model = "log_intensity"
    if os.path.exists(output_path):
        click.secho("Output path '%s' exists, removing..." % output_path, fg='yellow')
        os.remove(output_path)
    database_connection = DatabaseBoundOperation(database_connection)
    decoy_database_connection = DatabaseBoundOperation(decoy_database_connection)
    ms_data = ProcessedMSFileLoader(sample_path, use_index=False)
    try:
        sample_run = ms_data.sample_run
    except StopIteration:
        sample_run = Sample(name=os.path.basename(sample_path).split(".")[0], id=str(uuid4()))

    try:
        target_hypothesis = get_by_name_or_id(
            database_connection, GlycopeptideHypothesis, target_hypothesis_identifier)
    except Exception:
        click.secho("Could not locate Target Glycopeptide Hypothesis with identifier %r" %
                    target_hypothesis_identifier, fg='yellow')
        raise click.Abort()

    try:
        decoy_hypothesis = get_by_name_or_id(
            decoy_database_connection, GlycopeptideHypothesis, decoy_hypothesis_identifier)
    except Exception:
        click.secho("Could not locate Decoy Glycopeptide Hypothesis with identifier %r" %
                    decoy_hypothesis_identifier, fg='yellow')
        raise click.Abort()

    tandem_scoring_model, evaluation_kwargs = validate_glycopeptide_tandem_scoring_function(
        context, tandem_scoring_model)

    mass_shifts = [validate_mass_shift(mass_shift, multiplicity)
                   for mass_shift, multiplicity in mass_shifts]
    expanded = []
    expanded = MzMLGlycanChromatogramAnalyzer.expand_mass_shifts(
        dict(mass_shifts), crossproduct=False)
    mass_shifts = expanded

    if analysis_name is None:
        analysis_name = "%s @ %s" % (sample_run.name, target_hypothesis.name)

    analysis_name = validate_analysis_name(
        context, database_connection.session, analysis_name)

    click.secho("Preparing analysis of %s by %s" % (
        sample_run.name, target_hypothesis.name), fg='cyan')
    analyzer = MultipartGlycopeptideLCMSMSAnalyzer(
        database_connection._original_connection,
        decoy_database_connection._original_connection,
        target_hypothesis.id,
        decoy_hypothesis.id,
        sample_path,
        output_path,
        analysis_name=analysis_name,
        grouping_error_tolerance=grouping_error_tolerance,
        mass_error_tolerance=mass_error_tolerance,
        msn_mass_error_tolerance=msn_mass_error_tolerance,
        psm_fdr_threshold=psm_fdr_threshold,
        tandem_scoring_model=tandem_scoring_model,
        glycan_score_threshold=glycan_score_threshold,
        mass_shifts=mass_shifts,
        n_processes=processes,
        spectrum_batch_size=workload_size,
        maximum_mass=maximum_mass,
        probing_range_for_missing_precursors=isotope_probing_range,
        use_memory_database=memory_database_index,
        fdr_estimation_strategy=fdr_estimation_strategy,
        glycosylation_site_models_path=glycoproteome_smoothing_model,
        fragile_fucose=False,
        rare_signatures=rare_signatures,
        evaluation_kwargs=evaluation_kwargs,
        model_retention_time=retention_time_modeling,
        oxonium_threshold=oxonium_threshold,
        peptide_masses_per_scan=peptide_masses_per_scan
    )
    analyzer.display_header()
    result = analyzer.start()
    gps, unassigned, target_decoy_set = result[:3]
    if save_intermediate_results is not None:
        analyzer.log("Saving Intermediate Results")
        with open(save_intermediate_results, 'wb') as handle:
            pickle.dump((target_decoy_set, gps), handle)
    del gps
    del unassigned
    del target_decoy_set
    if export:
        for export_type in set(export):
            click.echo(fmt_msg("Handling Export: %s" % (export_type,)))
            if export_type == 'csv':
                from glycresoft.cli.export import glycopeptide_identification
                base = os.path.splitext(output_path)[0]
                export_path = "%s-glycopeptides.csv" % (base,)
                context.invoke(
                    glycopeptide_identification,
                    database_connection=output_path,
                    analysis_identifier=analyzer.analysis_id,
                    output_path=export_path)
            elif export_type == 'psm-csv':
                from glycresoft.cli.export import glycopeptide_spectrum_matches
                base = os.path.splitext(output_path)[0]
                export_path = "%s-glycopeptide-spectrum-matches.csv" % (base,)
                context.invoke(
                    glycopeptide_spectrum_matches,
                    database_connection=output_path,
                    analysis_identifier=analyzer.analysis_id,
                    output_path=export_path)
            elif export_type == 'html':
                from glycresoft.cli.export import glycopeptide_identification
                base = os.path.splitext(output_path)[0]
                export_path = "%s-report.html" % (base,)
                context.invoke(
                    glycopeptide_identification,
                    database_connection=output_path,
                    analysis_identifier=analyzer.analysis_id,
                    output_path=export_path,
                    report=True)



@analyze.command("fit-glycoproteome-smoothing-model", short_help=(
    "Fit a site-specific glycome network smoothing model for each site in the glycoproteome"))
@click.pass_context
@processes_option
@click.option("-i", "--analysis-path", nargs=2, multiple=True, required=True)
@click.option("-o", "--output-path", type=click.Path(writable=True), required=True)
@click.option('-q', '--fdr-threshold', type=float, default=0.05,
              help="The FDR threshold to apply when selecting identified glycopeptides")
@click.option("-P", "--glycopeptide-hypothesis", type=(DatabaseConnectionParam(exists=True), str))
@click.option("-g", "--glycan-hypothesis", type=(DatabaseConnectionParam(exists=True), str))
@click.option("-u", "--unobserved-penalty-scale", type=float, default=1.0, required=False,
              help="A penalty to scale unobserved-but-suggested glycans by. Defaults to 1.0, no penalty.")
@click.option("-a", "--smoothing-limit", type=float, default=0.2,
              help="An upper bound on the network smoothness to use when estimating the posterior probability.")
@click.option("-r", "--require-multiple-observations/--no-require-multiple-observations", is_flag=True, default=False,
              help=(
                  "Require a glycan/glycosite combination be observed in multiple samples to treat it as real."
                  " Defaults to False."))
@click.option("-w", "--network-path", type=click.Path(exists=True, file_okay=True, dir_okay=False), required=False,
              default=None,
              help=("The path to a text file defining the glycan network and its neighborhoods, as produced by "
                    "`glycresfoft build-hypothesis glycan-network`, otherwise the default human N-glycan network "
                    "will be used with the glycans defined in `-g`."))
@click.option("-D / -ND", "--include-decoys / --exclude-decoys", is_flag=True, default=True,
              help="Include decoy glycans in the network.", cls=HiddenOption)
def fit_glycoproteome_model(context, analysis_path, output_path, glycopeptide_hypothesis, glycan_hypothesis,
                            processes=4, unobserved_penalty_scale=None, smoothing_limit=0.2,
                            require_multiple_observations=True, fdr_threshold=0.05,
                            network_path=None, include_decoys=True):
    analysis_path_set = analysis_path
    analysis_path_set_transformed = []
    if require_multiple_observations and len(analysis_path_set) == 1:
        click.secho("Requested multiple observations required but only one analysis provided"
                    " discarding multiple observation requirement.", fg='yellow')
    click.echo(f"Collecting {len(analysis_path_set)} analyses")
    for analysis_path, analysis_id in analysis_path_set:
        database_connection = DatabaseBoundOperation(analysis_path)
        try:
            click.echo("Checking analysis %s:%s" %
                       (analysis_path, analysis_id))
            _analysis = get_by_name_or_id(database_connection, Analysis, analysis_id)
        except Exception:
            click.secho("Could not locate an Analysis in %r with identifier %r" %
                        (analysis_path, analysis_id), fg='yellow')
            raise click.Abort()
        # analysis_path_set_transformed.append((analysis_path, analysis.id))
        analysis_path_set_transformed.append((analysis_path, int(analysis_id)))
    analysis_path_set = analysis_path_set_transformed

    if network_path is not None:
        click.echo("Loading graph from \"%s\"" % (network_path, ))
        with open(network_path, 'r') as netfile:
            network = GraphReader(netfile).network
    else:
        network = None

    glycopeptide_database_connection_path, hypothesis_identifier = glycopeptide_hypothesis
    glycopeptide_database_connection = DatabaseBoundOperation(glycopeptide_database_connection_path)
    try:
        glycopeptide_hypothesis = get_by_name_or_id(
            glycopeptide_database_connection, GlycopeptideHypothesis, hypothesis_identifier)
    except Exception:
        click.secho("Could not locate a Glycopeptide Hypothesis with identifier %r" %
                    hypothesis_identifier, fg='yellow')
        raise click.Abort()

    glycan_database_connection_path, hypothesis_identifier = glycan_hypothesis
    glycan_database_connection = DatabaseBoundOperation(
        glycan_database_connection_path)
    try:
        glycan_hypothesis = get_by_name_or_id(
            glycan_database_connection, GlycanHypothesis, hypothesis_identifier)
    except Exception:
        click.secho("Could not locate a Glycopeptide Hypothesis with identifier %r" %
                    hypothesis_identifier, fg='yellow')
        raise click.Abort()
    workflow = site_model.GlycoproteinSiteModelBuildingWorkflow.from_paths(
        analysis_path_set, glycopeptide_database_connection_path, glycopeptide_hypothesis.id,
        glycan_database_connection_path, glycan_hypothesis.id, unobserved_penalty_scale, smoothing_limit,
        require_multiple_observations, output_path=output_path, n_threads=processes,
        q_value_threshold=fdr_threshold, network=network, include_decoy_glycans=False)
    workflow.start()


class RegularizationParameterType(click.ParamType):
    name = "\"grid\" or NUMBER > 0 or [grid|NUMBER]/NUMBER"
    seperator = '/'

    def convert(self, value, param, ctx):
        sep_count = value.count(self.seperator)
        if sep_count > 1:
            self.fail("regularization parameter split \"%s\" cannot be "
                      "used more than once" % self.seperator)
        elif sep_count == 1:
            parts = value.split(self.seperator)
            parts = (self.convert(parts[0], param, ctx),
                     self.convert(parts[1], param, ctx))
            if parts[1] == 'grid':
                self.fail("The second regularization parameter cannot be \"grid\"")
            return parts

        value = value.strip().lower()
        if value == 'grid':
            return LaplacianRegularizedChromatogramProcessor.GRID_SEARCH
        else:
            try:
                value = float(value)
                if value < 0:
                    self.fail("regularization parameter must be either \"grid\" or"
                              " a non-negative number between 0 and 1")
                return value
            except ValueError:
                self.fail("regularization parameter must be either \"grid\" or"
                          " a number between 0 and 1")


@analyze.command("search-glycan", short_help=('Search processed data for'
                                              ' glycan compositions'))
@click.pass_context
@database_connection_arg()
@sample_path_arg
@hypothesis_identifier_arg("glycan")
@click.option("-m", "--mass-error-tolerance", type=RelativeMassErrorParam(), default=1e-5,
              help=("Mass accuracy constraint, in parts-per-million "
                    "error, for matching."))
@click.option("-mn", "--msn-mass-error-tolerance", type=RelativeMassErrorParam(), default=2e-5,
              help="Mass accuracy constraint, in parts-per-million error, for matching MS^n ions.")
@click.option("-g", "--grouping-error-tolerance", type=RelativeMassErrorParam(), default=1.5e-5,
              help=("Mass accuracy constraint, in parts-per-million error, for"
                    " grouping chromatograms."))
@click.option("-n", "--analysis-name", default=None,
              help='Name for analysis to be performed.')
@click.option("-a", "--mass_shift", 'mass_shifts', multiple=True, nargs=2,
              help=("Adducts to consider. Specify name or formula, and a"
                    " multiplicity."))
@click.option("--mass_shift-combination-limit", type=int, default=8, help=(
    "Maximum number of mass_shift combinations to consider"))
@click.option("-d", "--minimum-mass", default=500., type=float,
              help="The minimum mass to consider signal at.")
@click.option("-o", "--output-path", default=None, help=("Path to write resulting analysis to."),
              required=True)
@click.option("--interact", is_flag=True, cls=HiddenOption)
@click.option("-f", "--ms1-scoring-feature", "scoring_model_features", multiple=True,
              type=click.Choice(ms1_model_features.keys(lambda x: x.replace("_", "-"))),
              help="Additional features to include in evaluating chromatograms")
@click.option("-r", "--regularize", type=RegularizationParameterType(),
              help=("Apply Laplacian regularization with either a"
                    " specified weight or \"grid\" to grid search, or a pair of values"
                    " separated by a / to specify a weight or grid search "
                    " for model fitting and a separate weight for scoring"))
@click.option("-w", "--regularization-model-path", type=click.Path(exists=True),
              default=None,
              help="Path to a file containing a neighborhood model for regularization")
@click.option("-k", "--network-path", type=click.Path(exists=True), default=None,
              help=("Path to a file containing the glycan composition network"
                    " and neighborhood rules"))
@click.option("-t", "--delta-rt", default=0.5, type=float,
              help='The maximum time between observed data points before splitting features')
@click.option("--export", type=click.Choice(['csv', 'glycan-list', 'html', "model"]), multiple=True,
              help="export command to after search is complete")
@click.option('-s', '--require-msms-signature', type=float, default=0.0,
              help="Minimum oxonium ion signature required in MS/MS scans to include.")
@processes_option
def search_glycan(context, database_connection, sample_path,
                  hypothesis_identifier,
                  analysis_name, mass_shifts, grouping_error_tolerance=1.5e-5,
                  mass_error_tolerance=1e-5, minimum_mass=500.,
                  scoring_model=None, regularize=None, regularization_model_path=None,
                  network_path=None,
                  output_path=None, scoring_model_features=None,
                  delta_rt=0.5, export=None, interact=False,
                  require_msms_signature=0.0, msn_mass_error_tolerance=2e-5,
                  mass_shift_combination_limit=None,
                  processes=4):
    """Identify glycan compositions from preprocessed LC-MS data, stored in mzML
    format.
    """
    if os.path.exists(output_path):
        click.secho("Output path '%s' exists, removing..." % output_path, fg='yellow')
        os.remove(output_path)

    if scoring_model is None:
        scoring_model = GeneralScorer

    if mass_shift_combination_limit is None:
        mass_shift_combination_limit = 8

    if scoring_model_features:
        for feature in scoring_model_features:
            scoring_model.add_feature(validate_ms1_feature_name(feature))

    if regularization_model_path is not None:
        with open(regularization_model_path, 'r') as mod_file:
            regularization_model = GridPointSolution.load(mod_file)
    else:
        regularization_model = None

    if network_path is not None:
        with open(network_path, 'r') as netfile:
            network = GraphReader(netfile).network
    else:
        network = None

    database_connection = DatabaseBoundOperation(database_connection)
    ms_data = ProcessedMSFileLoader(sample_path, use_index=False)
    try:
        sample_run = ms_data.sample_run
    except StopIteration:
        sample_run = Sample(name=os.path.basename(sample_path).split(".")[0], id=str(uuid4()))

    try:
        hypothesis = get_by_name_or_id(
            database_connection, GlycanHypothesis, hypothesis_identifier)
    except Exception:
        click.secho("Could not locate a Glycan Hypothesis with identifier %r" %
                    hypothesis_identifier, fg='yellow')
        raise click.Abort()

    if analysis_name is None:
        analysis_name = "%s @ %s" % (sample_run.name, hypothesis.name)

    analysis_name = validate_analysis_name(
        context, database_connection.session, analysis_name)

    mass_shifts = [validate_mass_shift(mass_shift, multiplicity)
                   for mass_shift, multiplicity in mass_shifts]
    expanded = []
    expanded = MzMLGlycanChromatogramAnalyzer.expand_mass_shifts(dict(mass_shifts), limit=mass_shift_combination_limit)
    mass_shifts = expanded

    click.secho("Preparing analysis of %s by %s" %
                (sample_run.name, hypothesis.name), fg='cyan')

    analyzer = MzMLGlycanChromatogramAnalyzer(
        database_connection._original_connection, hypothesis.id,
        sample_path=sample_path, output_path=output_path, mass_shifts=mass_shifts,
        mass_error_tolerance=mass_error_tolerance,
        msn_mass_error_tolerance=msn_mass_error_tolerance,
        grouping_error_tolerance=grouping_error_tolerance,
        scoring_model=scoring_model,
        minimum_mass=minimum_mass,
        regularize=regularize,
        regularization_model=regularization_model,
        network=network,
        analysis_name=analysis_name,
        delta_rt=delta_rt,
        require_msms_signature=require_msms_signature,
        n_processes=processes)
    analyzer.display_header()
    analyzer.start()
    if interact:
        try:
            import IPython
            click.secho(fmt_msg("Beginning Interactive Session..."), fg='cyan')
            IPython.embed()
        except ImportError:
            click.secho(fmt_msg("Interactive Session Not Supported"), fg='red')
    if export:
        for export_type in set(export):
            click.echo(fmt_msg("Handling Export: %s" % (export_type,)))
            if export_type == 'csv':
                from glycresoft.cli.export import glycan_composition_identification
                base = os.path.splitext(output_path)[0]
                export_path = "%s-glycan-chromatograms.csv" % (base,)
                context.invoke(
                    glycan_composition_identification,
                    database_connection=output_path,
                    analysis_identifier=analyzer.analysis_id,
                    output_path=export_path)
            elif export_type == 'html':
                from glycresoft.cli.export import glycan_composition_identification
                base = os.path.splitext(output_path)[0]
                export_path = "%s-report.html" % (base,)
                context.invoke(
                    glycan_composition_identification,
                    database_connection=output_path,
                    analysis_identifier=analyzer.analysis_id,
                    output_path=export_path,
                    report=True)
            elif export_type == 'glycan-list':
                from glycresoft.cli.export import glycan_hypothesis
                base = os.path.splitext(output_path)[0]
                export_path = "%s-glycan-chromatograms.csv" % (base,)
                context.invoke(
                    glycan_hypothesis,
                    database_connection=output_path,
                    hypothesis_identifier=analyzer.hypothesis_id,
                    output_path=export_path,
                    importable=True)
            elif export_type == "model":
                base = os.path.splitext(output_path)[0]
                export_path = "%s-regularization-parameters.txt" % (base,)
                params = analyzer.analysis.parameters.get("network_parameters")
                if params is None:
                    click.secho("No parameters were fitted, skipping \"model\"", fg='red')
                else:
                    with open(export_path, 'w') as fp:
                        params.dump(fp)
            else:
                click.secho("Unrecognized Export: %s" % (export_type,), fg='yellow')


@analyze.command("summarize-chromatograms",
                 short_help="Simply summarize coherent chromatograms by mass and signal with time boundaries")
@click.pass_context
@sample_path_arg
@click.option("-o", "--output-path", default=None, type=click.Path(writable=True),
              help=("Path to write resulting analysis to as a CSV"))
@click.option("-e", "--evaluate", is_flag=True,
              help=("Should all chromatograms be evaluated. Can greatly increase runtime."))
def summarize_chromatograms(context, sample_path, output_path, evaluate=False):
    task = ChromatogramSummarizer(sample_path, evaluate=evaluate)
    chromatograms, summary_chromatograms = task.start()
    if output_path is None:
        output_path = os.path.splitext(sample_path)[0] + '.chromatograms.csv'
    from glycresoft.output import (
        SimpleChromatogramCSVSerializer, SimpleScoredChromatogramCSVSerializer)
    click.echo("Writing chromatogram summaries to %s" % (output_path, ))
    with open(output_path, 'wb') as fh:
        if evaluate:
            writer = SimpleScoredChromatogramCSVSerializer(fh, chromatograms)
        else:
            writer = SimpleChromatogramCSVSerializer(fh, chromatograms)
        writer.run()



@analyze.group(short_help="Model and predict the retention time of glycopeptides")
def retention_time():
    pass


@retention_time.command("fit-glycopeptide-retention-time", short_help="Model the retention time of glycopeptides")
@click.pass_context
@click.argument("chromatogram-csv-path", type=click.Path(readable=True))
@click.argument("output-path", type=click.Path(writable=True))
@click.option("-t", "--test-chromatogram-csv-path", type=click.Path(readable=True),
              help=("Path to glycopeptide CSV chromatograms to evaluate with the model, but not fit on"))
@click.option("-p", "--prefer-joint-model", is_flag=True,
              help="Prefer the joint model over the peptide-specific ones")
@click.option('-m', '--minimum-observations-for-specific-model', type=int, default=20,
              help="The minimum number of observations required to fit a peptide sequence-specific model")
@click.option('-r', '--use-retention-time-normalization', is_flag=True)
def glycopeptide_retention_time_fit(context, chromatogram_csv_path, output_path=None,
                                    test_chromatogram_csv_path=None, prefer_joint_model=False,
                                    minimum_observations_for_specific_model=20, use_retention_time_normalization=False):
    '''Fit a retention time model to glycopeptide chromatogram observations, and evaluate the chromatographic apex
    retention of each glycopeptide based upon the model fit. For large enough groups, try to fit a model to just
    that group.
    '''
    with open(chromatogram_csv_path, 'rt') as fh:
        chromatograms = GlycopeptideChromatogramProxy.from_csv(fh)
    if test_chromatogram_csv_path is not None:
        with open(test_chromatogram_csv_path, 'rt') as fh:
            test_chromatograms = GlycopeptideChromatogramProxy.from_csv(fh)
    else:
        test_chromatograms = None
    modeler = GlycopeptideElutionTimeModeler(
        chromatograms, test_chromatograms=test_chromatograms,
        prefer_joint_model=prefer_joint_model,
        use_retention_time_normalization=use_retention_time_normalization,
        minimum_observations_for_specific_model=minimum_observations_for_specific_model)
    modeler.run()
    if output_path:
        modeler.write(output_path)


@retention_time.command(
    "evaluate-glycopeptide-retention-time",
    short_help="Predict and score the retention time of glycopeptides using an existing model")
@click.pass_context
@click.argument("model-path", type=click.Path(readable=True))
@click.argument("chromatogram-csv-path", type=click.Path(readable=True))
@click.argument("output-path", type=click.Path(writable=True))
def glycopeptide_retention_time_predict(context, model_path, chromatogram_csv_path, output_path):
    """Predict and evaluate the retention time of glycopeptides using an existing model.

    The model file should have the extension .pkl and have been produced by the same version of GlycReSoft.
    """
    with open(model_path, 'rb') as fh:
        modeler = pickle.load(fh)
    with open(chromatogram_csv_path, 'rt') as fh:
        chromatograms = GlycopeptideChromatogramProxy.from_csv(fh)
    modeler.evaluate(chromatograms)
    with open(output_path, 'wt') as fh:
        GlycopeptideChromatogramProxy.to_csv(chromatograms, fh)


@analyze.command("summarize-analysis", short_help="Briefly summarize the results of an analysis output file")
@click.pass_context
@database_connection_arg()
@analysis_identifier_arg("glycopeptide")
def summarize_analysis(context: click.Context, database_connection, analysis_identifier, verbose=False):
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable

    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        context.abort()

    ads = serialize.AnalysisDeserializer(database_connection._original_connection, analysis_id=analysis.id)

    idgps = ads.query(serialize.IdentifiedGlycopeptide).all()
    idgp_05 = 0
    idgp_01 = 0

    for idgp in idgps:
        q = idgp.q_value
        if q <= 0.05:
            idgp_05 += 1
        if q <= 0.01:
            idgp_01 += 1

    gpsms = ads.get_glycopeptide_spectrum_matches(0.05).all()
    gpsm_05 = 0
    gpsm_01 = 0

    seen_spectrum_05 = set()
    seen_spectrum_01 = set()
    for gpsm in gpsms:
        scan_id = gpsm.scan.scan_id
        q = gpsm.q_value
        if q <= 0.05:
            if scan_id not in seen_spectrum_05:
                seen_spectrum_05.add(scan_id)
                gpsm_05 += 1
        if q <= 0.01:
            if scan_id not in seen_spectrum_01:
                seen_spectrum_01.add(scan_id)
                gpsm_01 += 1

    click.echo(f"Name: {ads.analysis.name}")
    click.echo(f"Identified Glycopeptides @ 5% FDR: {idgp_05}")
    click.echo(f"Identified Glycopeptides @ 1% FDR: {idgp_01}")

    click.echo(f"Identified GPSMs @ 5% FDR: {gpsm_05}")
    click.echo(f"Identified GPSMs @ 1% FDR: {gpsm_01}")



import os
import operator
import logging
import struct
import ctypes

from typing import Dict, List, Optional, Set

from collections import namedtuple, defaultdict

from glycopeptidepy import PeptideSequence, GlycosylationType
from glycopeptidepy.structure.glycan import GlycanCompositionWithOffsetProxy
from glycopeptidepy.structure.sequence import (
    _n_glycosylation,
    _o_glycosylation,
    _gag_linker_glycosylation,
)

from glypy.structure.glycan_composition import FrozenMonosaccharideResidue
from glypy.utils.enum import EnumValue

from ms_deisotope import isotopic_shift
from ms_deisotope.data_source import ProcessedScan

from glycresoft.chromatogram_tree.mass_shift import MassShift
from glycresoft.serialize.hypothesis.glycan import GlycanTypes

from glycresoft.task import LoggingMixin

from glycresoft.chromatogram_tree import Unmodified

from glycresoft.structure.lru import LRUMapping
from glycresoft.structure import (
    FragmentCachingGlycopeptide,
    DecoyFragmentCachingGlycopeptide,
    PeptideProteinRelation,
)
from glycresoft.structure.structure_loader import (
    CachingStubGlycopeptideStrategy,
    PeptideDatabaseRecord,
)

from glycresoft.tandem.oxonium_ions import (
    OxoniumIndex,
    SignatureIonIndex,
    SignatureIonIndexMatch,
    gscore_scanner,
)

from glycresoft.composition_distribution_model.site_model import GlycoproteomePriorAnnotator

from glycresoft.database import intervals
from glycresoft.database.mass_collection import NeutralMassDatabase, SearchableMassCollection
from glycresoft.database.builder.glycopeptide.common import limiting_combinations

from glycresoft.structure.enums import SpectrumMatchClassification as StructureClassification

from ...workload import WorkloadManager
from ..core_search import GlycanCombinationRecord, IndexedGlycanFilteringPeptideMassEstimator


logger = logging.Logger("glycresoft.dynamic_generation")


debug_mode = bool(os.environ.get("GLYCRESOFTDEBUG"))


_glycopeptide_key_t = namedtuple(
    "glycopeptide_key",
    (
        "start_position",
        "end_position",
        "peptide_id",
        "protein_id",
        "hypothesis_id",
        "glycan_combination_id",
        "structure_type",
        "site_combination_index",
        "glycosylation_type",
    ),
)


class glycopeptide_key_t_base(_glycopeptide_key_t):
    __slot__ = ()

    def copy(self, structure_type=None):
        if structure_type is None:
            structure_type = self.structure_type
        return self._replace(structure_type=structure_type)

    def as_dict(self, stringify=False):
        d = {}
        for i, label in enumerate(self._fields):
            d[label] = str(self[i]) if stringify else self[i]
        return d


MonosaccharideResidues = Set[FrozenMonosaccharideResidue]
GlycanCombinationRecords = List[GlycanCombinationRecord]

# try:
#     from glycresoft._c.structure.structure_loader import glycopeptide_key as glycopeptide_key_t_base
# except ImportError:
#     pass

# The site_combination_index slot is necessary to distinguish alternative arrangements of
# the same combination of glycans on the same amino acid sequence. The placeholder value
# used for unassigned glycosite combination permutations, the maximum value that fits in
# an unsigned 4 byte integer (L signifier in struct spec). This means defines the upper
# limit on the number of distinct combinations that can be uniquely addressed is
# 2 ** 32 - 1 (4294967295).
#
# Note that the glycoform generators all use (circa 2018) :func:`~.limiting_combinations`,
# which only produces the first 100 iterations to avoid producing too many permutations of
# O-glycans.
placeholder_permutation = ctypes.c_uint32(-1).value

TT = StructureClassification.target_peptide_target_glycan


class glycopeptide_key_t(glycopeptide_key_t_base):
    __slots__ = ()

    start_position: int
    end_position: int
    peptide_id: int
    protein_id: int
    hypothesis_id: int
    glycan_combination_id: int
    structure_type: EnumValue
    site_combination_index: int
    glycosylation_type: str

    struct_spec = struct.Struct("!LLLLLLLL")

    def serialize(self):
        return self.struct_spec.pack(*self)

    @classmethod
    def parse(cls, binary):
        return cls(*cls.struct_spec.unpack(binary))

    def to_decoy_glycan(self) -> "glycopeptide_key_t":
        structure_type = self.structure_type
        new_tp = StructureClassification[
            structure_type | StructureClassification.target_peptide_decoy_glycan
        ]
        return self.copy(new_tp)

    def is_decoy(self) -> bool:
        return self.structure_type != 0


CORE_TO_GLYCOSYLATION_TYPE = {
    _n_glycosylation: GlycanTypes.n_glycan,
    _o_glycosylation: GlycanTypes.o_glycan,
    _gag_linker_glycosylation: GlycanTypes.gag_linker,
}


class GlycoformGeneratorBase(LoggingMixin):
    glycan_combinations: NeutralMassDatabase[GlycanCombinationRecord]
    _peptide_cache: LRUMapping
    default_structure_type: StructureClassification
    glycan_prior_model: GlycoproteomePriorAnnotator
    expand_combinatorics: bool

    @classmethod
    def from_hypothesis(cls, session, hypothesis_id, **kwargs):
        """Build a glycan combination index from a :class:`~.GlycanHypothesis`

        Parameters
        ----------
        session: :class:`DatabaseBoundOperation`
            The database connection to use to load glycan combinations.
        hypothesis_id: int
            The id key of the :class:`~.GlycanHypothesis` to build from

        Returns
        -------
        :class:`GlycoformGeneratorBase`
        """
        glycan_combinations = GlycanCombinationRecord.from_hypothesis(session, hypothesis_id)
        return cls(glycan_combinations, **kwargs)

    def __init__(
        self,
        glycan_combinations,
        cache_size=None,
        default_structure_type=TT,
        expand_combinatorics: bool = True,
        *args,
        **kwargs
    ):
        if not isinstance(glycan_combinations, SearchableMassCollection):
            glycan_combinations = NeutralMassDatabase(
                list(glycan_combinations), operator.attrgetter("dehydrated_mass")
            )
        if cache_size is None:
            cache_size = 2**15
        else:
            cache_size = int(cache_size)
        self.glycan_combinations = glycan_combinations
        self._peptide_cache = LRUMapping(cache_size)
        self._cache_hit = 0
        self._cache_miss = 0
        self.default_structure_type = default_structure_type
        self.glycan_prior_model = kwargs.pop("glycan_prior_model", None)
        self.expand_combinatorics = expand_combinatorics
        super(GlycoformGeneratorBase, self).__init__(*args, **kwargs)

    def handle_glycan_combination(
        self,
        peptide_obj: PeptideSequence,
        peptide_record: PeptideDatabaseRecord,
        glycan_combination: GlycanCombinationRecord,
        glycosylation_sites: List[int],
        core_type: GlycosylationType,
    ) -> List["Record"]:
        glycosylation_type = CORE_TO_GLYCOSYLATION_TYPE[core_type]
        key = self._make_key(peptide_record, glycan_combination, glycosylation_type)
        if key in self._peptide_cache:
            self._cache_hit += 1
            return self._peptide_cache[key]
        self._cache_miss += 1
        protein_relation = PeptideProteinRelation(
            key.start_position, key.end_position, key.protein_id, key.hypothesis_id
        )
        glycosylation_sites_unoccupied = set(glycosylation_sites)
        for site in list(glycosylation_sites_unoccupied):
            if peptide_obj[site][1]:
                glycosylation_sites_unoccupied.remove(site)

        site_combinations = list(
            limiting_combinations(
                glycosylation_sites_unoccupied,
                glycan_combination.count,
                100 if self.expand_combinatorics else 1
            )
        )

        result_set = [None for i in site_combinations]
        for i, site_set in enumerate(site_combinations):
            glycoform = peptide_obj.clone(share_cache=False)
            glycoform.id = key._replace(site_combination_index=i)
            glycoform.glycan = GlycanCompositionWithOffsetProxy(glycan_combination.composition)
            for site in site_set:
                glycoform.add_modification(site, core_type.name)
            glycoform.protein_relation = protein_relation
            if self.glycan_prior_model is not None:
                glycoform.glycan_prior = self.glycan_prior_model.score(
                    glycoform, key.structure_type
                )
            result_set[i] = glycoform
        result_set = Record.build(result_set)
        self._peptide_cache[key] = result_set
        return result_set

    def _make_key(
        self,
        peptide_record: PeptideDatabaseRecord,
        glycan_combination: GlycanCombinationRecord,
        glycosylation_type: str,
        structure_type: StructureClassification = None,
    ) -> glycopeptide_key_t:
        if structure_type is None:
            structure_type = self.default_structure_type
        key = glycopeptide_key_t(
            peptide_record.start_position,
            peptide_record.end_position,
            peptide_record.id,
            peptide_record.protein_id,
            peptide_record.hypothesis_id,
            glycan_combination.id,
            structure_type,
            placeholder_permutation,
            glycosylation_type,
        )
        return key

    def handle_n_glycan(self, peptide_obj, peptide_record, glycan_combination):
        if glycan_combination.size < 5:
            return []
        return self.handle_glycan_combination(
            peptide_obj,
            peptide_record,
            glycan_combination,
            peptide_record.n_glycosylation_sites,
            _n_glycosylation,
        )

    def handle_o_glycan(self, peptide_obj, peptide_record, glycan_combination):
        return self.handle_glycan_combination(
            peptide_obj,
            peptide_record,
            glycan_combination,
            peptide_record.o_glycosylation_sites,
            _o_glycosylation,
        )

    def handle_gag_linker(self, peptide_obj, peptide_record, glycan_combination):
        return self.handle_glycan_combination(
            peptide_obj,
            peptide_record,
            glycan_combination,
            peptide_record.gagylation_sites,
            _gag_linker_glycosylation,
        )

    def reset_cache(self, **kwargs):
        self._cache_hit = 0
        self._cache_miss = 0
        self._peptide_cache.clear()

    def reset(self, **kwargs):
        self.reset_cache(**kwargs)


class PeptideGlycosylator(GlycoformGeneratorBase):
    peptides: NeutralMassDatabase
    peptide_to_group_id: Dict[int, List[int]]

    def __init__(
        self,
        peptide_records,
        glycan_combinations,
        cache_size=2**16,
        default_structure_type=TT,
        *args,
        **kwargs
    ):
        super(PeptideGlycosylator, self).__init__(
            glycan_combinations,
            default_structure_type=default_structure_type,
            cache_size=cache_size,
            *args,
            **kwargs
        )
        if not isinstance(peptide_records, SearchableMassCollection):
            peptide_records = NeutralMassDatabase(peptide_records)
        self.peptides = peptide_records
        self.peptide_to_group_id = None

    def build_peptide_groups(self):
        peptide_groups = defaultdict(list)
        for peptide in self.peptides:
            peptide_groups[peptide.modified_peptide_sequence].append(peptide.id)

        sequence_to_group_id = {}
        for i, key in enumerate(peptide_groups):
            sequence_to_group_id[key] = i

        peptide_to_group_id = {}
        for peptide in self.peptides:
            peptide_to_group_id[peptide.id] = sequence_to_group_id[
                peptide.modified_peptide_sequence
            ]
        sequence_to_group_id.clear()
        peptide_groups.clear()
        self.peptide_to_group_id = peptide_to_group_id

    def filter_glycan_proposals_by_signature_ions(
        self,
        glycan_combinations: GlycanCombinationRecords,
        signature_ion_match_index: SignatureIonIndexMatch,
    ) -> GlycanCombinationRecords:
        """Remove any glycans for which an expected signature ion is absent (or where the best matching peak is
        less than 1% of the base peak). Requires the signature ion index is initialized.
        """
        result = []
        for glycan in glycan_combinations:
            rec = signature_ion_match_index.record_for(glycan.composition)
            if rec.expected_matches:
                miss = 0
                for _sig, peak in rec.expected_matches.items():
                    if peak is None:
                        miss += 1
                    elif peak.intensity / signature_ion_match_index.base_peak_intensity < 0.01:
                        miss += 1
                if miss:
                    continue
            result.append(glycan)
        return result

    def handle_peptide_mass(
        self,
        peptide_mass: float,
        intact_mass: float,
        error_tolerance: float = 1e-5,
        signature_ion_match_index: SignatureIonIndexMatch = None,
    ) -> List["Record"]:
        peptide_records = self.peptides.search_mass_ppm(peptide_mass, error_tolerance)
        glycan_mass = intact_mass - peptide_mass
        glycan_combinations = self.glycan_combinations.search_mass_ppm(glycan_mass, error_tolerance)
        if signature_ion_match_index:
            glycan_combinations = self.filter_glycan_proposals_by_signature_ions(
                glycan_combinations, signature_ion_match_index
            )
        result_set = []
        for peptide in peptide_records:
            self._combinate(peptide, glycan_combinations, result_set)
        return result_set

    def _combinate(
        self,
        peptide: PeptideDatabaseRecord,
        glycan_combinations: GlycanCombinationRecords,
        result_set: Optional[List["Record"]] = None,
    ) -> List["Record"]:
        if result_set is None:
            result_set = []
        peptide_obj = peptide.convert()
        for glycan_combination in glycan_combinations:
            for tp in glycan_combination.glycan_types:
                tp = GlycosylationType[tp]
                if tp is GlycosylationType.n_linked:
                    result_set.extend(
                        self.handle_n_glycan(peptide_obj.clone(), peptide, glycan_combination)
                    )
                elif tp is GlycosylationType.o_linked:
                    result_set.extend(
                        self.handle_o_glycan(peptide_obj.clone(), peptide, glycan_combination)
                    )
                elif tp is GlycosylationType.glycosaminoglycan:
                    result_set.extend(
                        self.handle_gag_linker(peptide_obj.clone(), peptide, glycan_combination)
                    )
        return result_set

    def generate_crossproduct(self, lower_bound=0, upper_bound=float("inf")):
        minimum_peptide_mass = max(lower_bound - self.glycan_combinations.highest_mass, 0)
        maximum_peptide_mass = max(upper_bound - self.glycan_combinations.lowest_mass, 0)
        peptides = self.peptides.search_between(
            max(minimum_peptide_mass - 10, 0), maximum_peptide_mass
        )
        for peptide in peptides:
            if not peptide.has_glycosylation_sites():
                continue
            glycan_mass_limit = upper_bound - peptide.calculated_mass
            if glycan_mass_limit < 0:
                continue
            minimum_glycan_mass = max(lower_bound - peptide.calculated_mass, 0)
            glycan_combinations = self.glycan_combinations.search_between(
                minimum_glycan_mass, glycan_mass_limit + 10
            )
            for solution in self._combinate(peptide, glycan_combinations):
                total_mass = solution.total_mass
                if total_mass < lower_bound or total_mass > upper_bound:
                    continue
                yield solution

    def reset(self, **kwargs):
        super(PeptideGlycosylator, self).reset(**kwargs)
        self.peptides.reset(**kwargs)


class DynamicGlycopeptideSearchBase(LoggingMixin):
    neutron_shift = isotopic_shift()

    def handle_scan_group(
        self,
        group: List,
        precursor_error_tolerance: float = 1e-5,
        mass_shifts: Optional[List[MassShift]] = None,
    ):
        raise NotImplementedError()

    def reset(self, **kwargs):
        self.peptide_glycosylator.reset(**kwargs)

    def construct_peptide_groups(self, workload: WorkloadManager):
        if self.peptide_glycosylator.peptide_to_group_id is None:
            self.peptide_glycosylator.build_peptide_groups()
        workload.hit_group_map.clear()
        for hit in workload.hit_map.values():
            workload.hit_group_map[
                self.peptide_glycosylator.peptide_to_group_id[hit.id.peptide_id]
            ].add(hit.id)


class PeptideMassFilterBase:
    peptide_glycosylator: PeptideGlycosylator
    product_error_tolerance: float
    glycan_score_threshold: float
    min_fragments: int
    peptide_mass_predictor: IndexedGlycanFilteringPeptideMassEstimator
    peptide_masses_per_scan: int
    probing_range_for_missing_precursors: int
    trust_precursor_fits: bool
    oxonium_ion_index: OxoniumIndex
    signature_ion_index: SignatureIonIndex
    oxonium_ion_threshold: float

    monosaccharides: MonosaccharideResidues

    neutron_shift: float = isotopic_shift()

    def __init__(
        self,
        glycan_compositions: GlycanCombinationRecords,
        product_error_tolerance: float=2e-5,
        glycan_score_threshold: float=0.1,
        min_fragments: int=2,
        peptide_masses_per_scan: int=100,
        probing_range_for_missing_precursors: int=3,
        trust_precursor_fits: bool=True,
        fragment_weight: float=0.56,
        core_weight: float = 1.42,
        oxonium_ion_index: Optional[OxoniumIndex]=None,
        signature_ion_index: Optional[SignatureIonIndex]=None,
        oxonium_ion_threshold: float = 0.05
    ):
        # Intentionally use a larger core_weight here than in the real scoring function to
        # prefer solutions with more core fragments, but not to discard them later?
        if min_fragments is None:
            min_fragments = 2
        self.product_error_tolerance = product_error_tolerance
        self.glycan_score_threshold = glycan_score_threshold
        self.min_fragments = int(min_fragments)
        self.peptide_mass_predictor = IndexedGlycanFilteringPeptideMassEstimator(
            glycan_compositions,
            product_error_tolerance=product_error_tolerance,
            fragment_weight=fragment_weight,
            core_weight=core_weight,
        )
        self.monosaccharides = self._monosaccharides_from_records(glycan_compositions)
        self.peptide_masses_per_scan = peptide_masses_per_scan
        self.probing_range_for_missing_precursors = probing_range_for_missing_precursors
        self.trust_precursor_fits = trust_precursor_fits
        self.oxonium_ion_index = oxonium_ion_index
        self.signature_ion_index = signature_ion_index
        self.oxonium_ion_threshold = oxonium_ion_threshold

    def _monosaccharides_from_records(self, glycan_combinations: GlycanCombinationRecords) -> MonosaccharideResidues:
        residues = set()
        for rec in glycan_combinations:
            residues.update(rec.composition)
        return {FrozenMonosaccharideResidue.from_iupac_lite(str(x)) for x in residues}


class PredictiveGlycopeptideSearch(PeptideMassFilterBase, DynamicGlycopeptideSearchBase):
    peptide_glycosylator: PeptideGlycosylator

    def __init__(
        self,
        peptide_glycosylator,
        product_error_tolerance=2e-5,
        glycan_score_threshold=0.1,
        min_fragments=2,
        peptide_masses_per_scan=100,
        probing_range_for_missing_precursors=3,
        trust_precursor_fits=True,
        fragment_weight=0.56,
        core_weight=1.42,
        oxonium_ion_index=None,
        signature_ion_index=None,
        oxonium_ion_threshold: float=0.05
    ):
        self.peptide_glycosylator = peptide_glycosylator
        super().__init__(
            self.peptide_glycosylator.glycan_combinations,
            product_error_tolerance=product_error_tolerance,
            glycan_score_threshold=glycan_score_threshold,
            min_fragments=min_fragments,
            peptide_masses_per_scan=peptide_masses_per_scan,
            probing_range_for_missing_precursors=probing_range_for_missing_precursors,
            trust_precursor_fits=trust_precursor_fits,
            fragment_weight=fragment_weight,
            core_weight=core_weight,
            oxonium_ion_index=oxonium_ion_index,
            signature_ion_index=signature_ion_index,
            oxonium_ion_threshold=oxonium_ion_threshold)

    def handle_scan_group(
        self,
        group: List[ProcessedScan],
        precursor_error_tolerance: float = 1e-5,
        mass_shifts: Optional[List[MassShift]] = None,
        workload: Optional[WorkloadManager] = None,
    ):
        if mass_shifts is None or not mass_shifts:
            mass_shifts = [Unmodified]
        if workload is None:
            workload = WorkloadManager()

        glycan_score_threshold = self.glycan_score_threshold
        min_fragments = self.min_fragments
        estimate_peptide_mass = self.peptide_mass_predictor.estimate_peptide_mass
        handle_peptide_mass = self.peptide_glycosylator.handle_peptide_mass
        peptide_masses_per_scan = self.peptide_masses_per_scan
        for scan in group:
            if gscore_scanner(scan) < self.oxonium_ion_threshold:
                continue
            workload.add_scan(scan)
            signature_ion_matches: SignatureIonIndexMatch = None

            if self.oxonium_ion_index:
                scan.annotations["oxonium_index_match"] = self.oxonium_ion_index.match(
                    scan.deconvoluted_peak_set, self.product_error_tolerance
                )
            if self.signature_ion_index:
                signature_ion_matches = scan.annotations[
                    "signature_index_match"
                ] = self.signature_ion_index.match(
                    scan.deconvoluted_peak_set, self.product_error_tolerance
                )
            n_glycopeptides = 0
            n_peptide_masses = 0
            if not scan.precursor_information.defaulted and self.trust_precursor_fits:
                probe = 0
            else:
                probe = self.probing_range_for_missing_precursors
            precursor_mass = scan.precursor_information.neutral_mass
            for i in range(probe + 1):
                neutron_shift = self.neutron_shift * i
                for mass_shift in mass_shifts:
                    seen = set()
                    mass_shift_name = mass_shift.name
                    mass_shift_mass = mass_shift.mass
                    intact_mass = precursor_mass - mass_shift_mass - neutron_shift
                    for peptide_mass_pred in estimate_peptide_mass(
                        scan,
                        topn=peptide_masses_per_scan,
                        mass_shift=mass_shift,
                        threshold=glycan_score_threshold,
                        min_fragments=min_fragments,
                        simplify=False,
                        query_mass=precursor_mass - neutron_shift,
                    ):
                        peptide_mass = peptide_mass_pred.peptide_mass
                        n_peptide_masses += 1
                        for candidate in handle_peptide_mass(
                            peptide_mass,
                            intact_mass,
                            self.product_error_tolerance,
                            signature_ion_match_index=signature_ion_matches,
                        ):
                            n_glycopeptides += 1
                            key = (candidate.id, mass_shift_name)
                            mass_threshold_passed = (
                                abs(intact_mass - candidate.total_mass) / intact_mass
                            ) <= precursor_error_tolerance
                            # What if it could be an N- or O-glycopeptide? Does this block the
                            # same glycan from being treated as both?
                            if key in seen:
                                continue
                            seen.add(key)
                            if mass_threshold_passed:
                                workload.add_scan_hit(scan, candidate, mass_shift_name)
        return workload


DEFAULT_THRESHOLD_CACHE_TOTAL_COUNT = 1e5


class IterativeGlycopeptideSearch(DynamicGlycopeptideSearchBase):
    total_mass_getter = operator.attrgetter("total_mass")

    def __init__(
        self,
        peptide_glycosylator,
        product_error_tolerance=2e-5,
        cache_size=2,
        threshold_cache_total_count=DEFAULT_THRESHOLD_CACHE_TOTAL_COUNT,
    ):
        self.peptide_glycosylator = peptide_glycosylator
        self.product_error_tolerance = product_error_tolerance
        self.interval_cache = intervals.LRUIntervalSet(max_size=cache_size)
        self.threshold_cache_total_count = threshold_cache_total_count

    def _make_new_glycopeptides_for_interval(self, lower_bound, upper_bound):
        generator = self.peptide_glycosylator.generate_crossproduct(lower_bound, upper_bound)
        generator = NeutralMassDatabase(list(generator), self.total_mass_getter)
        return intervals.ConcatenateMassIntervalNode(generator)

    def _upkeep_memory_intervals(self, lower_bound=0, upper_bound=float("inf")):
        """Perform routine maintainence of the interval cache, ensuring its size does not
        exceed the upper limit
        """
        n = len(self.interval_cache)
        if n > 1:
            while (
                len(self.interval_cache) > 1
                and self.interval_cache.total_count > self.threshold_cache_total_count
            ):
                logger.info(
                    "Upkeep Memory Intervals %d %d",
                    self.interval_cache.total_count,
                    len(self.interval_cache),
                )
                self.interval_cache.remove_lru_interval()
        n = len(self.interval_cache)
        if n == 1 and self.interval_cache.total_count > self.threshold_cache_total_count:
            segment = self.interval_cache[0]
            segment.constrain(lower_bound, upper_bound)

    def get_glycopeptides(self, lower_bound, upper_bound):
        self._upkeep_memory_intervals(lower_bound, upper_bound)
        q = intervals.QueryIntervalBase((lower_bound + upper_bound) / 2.0, lower_bound, upper_bound)
        match = self.interval_cache.find_interval(q)
        if match is not None:
            logger.debug("Nearest interval %r", match)
            # We are completely contained in an existing interval, so just
            # use that one.
            if match.contains_interval(q):
                logger.debug("Query interval %r was completely contained in %r", q, match)
                return match.group
            # We overlap with an extending interval, so we should populate
            # the new one and merge them.
            elif match.overlaps(q):
                q2 = q.difference(match)
                logger.debug(
                    "Query interval partially overlapped, creating disjoint interval %r", q2
                )
                match = self.interval_cache.extend_interval(
                    match, self._make_new_glycopeptides_for_interval(q2.start, q2.end)
                )
                return match.group
            # We might need to insert a new interval
            else:
                logger.debug("Query interval %r did not overlap with %r", q, match)
                return self._insert_interval(q.start, q.end)
        else:
            logger.debug("Query interval %r did not overlap with %r", q, match)
            return self._insert_interval(q.start, q.end)

    def _insert_interval(self, lower_bound, upper_bound):
        node = self._make_new_glycopeptides_for_interval(lower_bound, upper_bound)
        logger.debug("New Node: %r", node)
        # We won't insert this node if it is empty.
        if len(node.group) == 0:
            return node.group
        nearest_interval = self.interval_cache.find_interval(node)
        # Should an insert be performed if the query just didn't overlap well
        # with the database?
        if nearest_interval is None:
            # No nearby interval, so we should insert
            logger.debug("No nearby interval for %r", node)
            self.interval_cache.insert_interval(node)
            return node.group
        elif nearest_interval.overlaps(node):
            logger.debug("Nearest interval %r overlaps this %r", nearest_interval, node)
            nearest_interval = self.interval_cache.extend_interval(nearest_interval, node)
            return nearest_interval.group
        elif not nearest_interval.contains_interval(node):
            logger.debug("Nearest interval %r didn't contain this %r", nearest_interval, node)
            # Nearby interval didn't contain this interval
            self.interval_cache.insert_interval(node)
            return node.group
        else:
            # Situation unclear.
            # Not worth inserting, so just return the group
            logger.info("Unknown Condition Overlap %r / %r" % (node, nearest_interval))
            return nearest_interval.group

    def handle_scan_group(
        self, group, precursor_error_tolerance=1e-5, mass_shifts=None, workload=None
    ):
        if mass_shifts is None or not mass_shifts:
            mass_shifts = [Unmodified]
        if workload is None:
            workload = WorkloadManager()

        group = NeutralMassDatabase(group, lambda x: x.precursor_information.neutral_mass)
        lo_mass = group.lowest_mass
        hi_mass = group.highest_mass
        shifts = [m.mass for m in mass_shifts]

        lo_shift = min(shifts) - 0.5
        hi_shift = max(shifts) + 0.5

        lower_bound = lo_mass + lo_shift
        lower_bound = lower_bound - (lower_bound * precursor_error_tolerance)

        upper_bound = hi_mass + hi_shift
        upper_bound = upper_bound + (upper_bound * precursor_error_tolerance)

        id_to_scan = {}
        for scan in group:
            id_to_scan[scan.id] = scan
        logger.info("\nQuerying between %f and %f", lower_bound, upper_bound)
        candidates = self.get_glycopeptides(lower_bound, upper_bound)
        logger.info(
            "Interval from %f to %f contained %d candidates",
            lower_bound,
            upper_bound,
            len(candidates),
        )
        for scan in group:
            for mass_shift in mass_shifts:
                intact_mass = scan.precursor_information.neutral_mass - mass_shift.mass
                for candidate in candidates.search_mass_ppm(intact_mass, precursor_error_tolerance):
                    workload.add_scan_hit(scan, candidate, mass_shift.name)
        return workload


class CompoundGlycopeptideSearch(object):
    def __init__(self, glycopeptide_searchers=None):
        if glycopeptide_searchers is None:
            glycopeptide_searchers = []
        self.glycopeptide_searchers = list(glycopeptide_searchers)

    def add(self, glycopeptide_searcher):
        self.glycopeptide_searchers.append(glycopeptide_searcher)

    def handle_scan_group(
        self, group, precursor_error_tolerance=1e-5, mass_shifts=None, workload=None
    ):
        if mass_shifts is None:
            mass_shifts = [Unmodified]
        if workload is None:
            workload = WorkloadManager()
        for searcher in self.glycopeptide_searchers:
            searcher.handle_scan_group(
                group, precursor_error_tolerance, mass_shifts, workload=workload
            )
        return workload


class Record(object):
    __slots__ = ("glycopeptide", "id", "total_mass", "glycan_prior")

    glycopeptide: str
    id: glycopeptide_key_t
    total_mass: float
    glycan_prior: float

    def __init__(self, glycopeptide: FragmentCachingGlycopeptide = None):
        if glycopeptide is not None:
            self.glycopeptide = str(glycopeptide)
            self.id = glycopeptide.id
            self.total_mass = glycopeptide.total_mass
            self.glycan_prior = glycopeptide.glycan_prior
        else:
            self.glycopeptide = ""
            self.id = None
            self.total_mass = 0
            self.glycan_prior = 0.0

    def __repr__(self):
        return "Record(%s)" % self.glycopeptide

    def __getstate__(self):
        return (self.glycopeptide, self.id, self.total_mass, self.glycan_prior)

    def __setstate__(self, state):
        self.glycopeptide, self.id, self.total_mass, self.glycan_prior = state

    def __eq__(self, other):
        if other is None:
            return False
        return (self.glycopeptide == other.glycopeptide) and (self.id == other.id)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.id)

    def convert(self) -> FragmentCachingGlycopeptide:
        if self.id.structure_type.value & StructureClassification.target_peptide_decoy_glycan.value:
            structure = SharedCacheAwareDecoyFragmentCachingGlycopeptide(self.glycopeptide)
        else:
            structure = FragmentCachingGlycopeptide(self.glycopeptide)
        structure.id = self.id
        structure.glycan_id = self.id.glycan_combination_id
        structure.glycan_prior = self.glycan_prior
        return structure

    def serialize(self):
        id_bytes = self.id.serialize()
        mass_bytes = struct.pack("!d", self.total_mass)
        seq = (self.glycopeptide).encode("utf8")
        encoded = id_bytes + mass_bytes + seq
        return encoded

    @classmethod
    def parse(cls, bytestring):
        offset = glycopeptide_key_t.struct_spec.size
        rec_id = glycopeptide_key_t.parse(bytestring[:offset])
        mass = struct.unpack("!d", bytestring[offset : offset + 8])
        seq = bytestring[offset + 8 :].decode("utf8")
        inst = cls()
        inst.id = rec_id
        inst.total_mass = mass
        inst.glycopeptide = seq
        return inst

    @classmethod
    def build(cls, glycopeptides: List[FragmentCachingGlycopeptide]) -> List["Record"]:
        return [cls(p) for p in glycopeptides]

    def copy(self, structure_type=None) -> "Record":
        if structure_type is None:
            structure_type = self.id.structure_type
        inst = self.__class__()
        inst.id = self.id.copy(structure_type)
        inst.glycopeptide = self.glycopeptide
        inst.total_mass = self.total_mass
        return inst

    def to_decoy_glycan(self) -> "Record":
        inst = self.__class__()
        inst.id = self.id.to_decoy_glycan()
        inst.glycopeptide = self.glycopeptide
        inst.total_mass = self.total_mass
        return inst


class SharedCacheAwareDecoyFragmentCachingGlycopeptide(DecoyFragmentCachingGlycopeptide):
    def stub_fragments(self, *args, **kwargs):
        kwargs.setdefault("strategy", CachingStubGlycopeptideStrategy)
        key = self.fragment_caches.stub_fragment_key(self, args, kwargs)
        if key in self.fragment_caches:
            return self.fragment_caches[key]
        target_key = self.fragment_caches._make_target_key(key)
        if target_key in self.fragment_caches:
            result = self.fragment_caches[target_key]
            do_clone = True
        else:
            do_clone = False
            result = list(
                # Directly call the superclass method of FragmentCachingGlycopeptide as we
                # do not need to go through a preliminary round of cache key construction and
                # querying.
                super(
                    FragmentCachingGlycopeptide, self
                ).stub_fragments(  # pylint: disable=bad-super-call
                    *args, **kwargs
                )
            )
        result = self._permute_stub_masses(result, kwargs, do_clone=do_clone, min_shift_size=1)
        self.fragment_caches[key] = result
        return result


class Parser(object):
    def __init__(self, max_size=2**12, **kwargs):
        self.cache = LRUMapping(max_size)

    def __call__(self, record):
        key = record.id
        if key in self.cache:
            return self.cache[key]
        else:
            struct = record.convert()
            self.cache[key] = struct
            return struct


class IdKeyMaker(object):
    """Build an id-key structure for a glycopeptide with a new
    glycan composition given a glycopeptide of the same peptide
    backbone.
    """

    def __init__(self, valid_glycans):
        self.valid_glycans = valid_glycans
        self.lookup_map = {g: g.id for g in self.valid_glycans}

    def make_id_controlled_structures(self, structure, references):
        glycan_id = self.lookup_map[structure.glycan_composition]
        result = []
        for reference in references:
            ref_key = reference.id
            alt_key = glycopeptide_key_t(
                ref_key.start_position,
                ref_key.end_position,
                ref_key.peptide_id,
                ref_key.protein_id,
                ref_key.hypothesis_id,
                glycan_id,
                ref_key.structure_type,
                ref_key.site_combination_index,
                # TODO: Does this needs to be generic over all glycosylation types of
                # glycan_id if it could span multiple, or will we assume that all alternatives
                # will be in references?
                ref_key.glycosylation_type,
            )
            alt = structure.clone()
            alt.id = alt_key
            alt.protein_relation = reference.protein_relation
            result.append(alt)
        return result

    def __call__(self, structure, references):
        return self.make_id_controlled_structures(structure, references)



from typing import (Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union,
                    TYPE_CHECKING, NamedTuple)

import numpy as np

from ms_deisotope.data_source.scan import ProcessedScan
from ms_deisotope.data_source import ProcessedRandomAccessScanSource

from glycresoft.task import TaskBase

from glycresoft.chromatogram_tree.mass_shift import MassShiftBase
from glycresoft.scoring import ChromatogramSolution

from glycresoft.tandem.target_decoy.base import FDREstimatorBase

from ..spectrum_match import (
    SpectrumMatch,
    SpectrumSolutionSet,
    MultiScoreSpectrumMatch,
    SpectrumMatcherBase
)

from .base import Predicate, TargetType, SolutionEntry

if TYPE_CHECKING:
    from .chromatogram import (
        TandemAnnotatedChromatogram,
        SpectrumMatchSolutionCollectionBase,
        TandemSolutionsWithoutChromatogram
    )
    from glycresoft.scoring.elution_time_grouping.structure import (
        ChromatogramProxy,
        GlycopeptideChromatogramProxy
    )
    from glycresoft.scoring.elution_time_grouping.model import ModelEnsemble as RetentionTimeModelEnsemble


class MS2RevisionValidator(TaskBase):
    threshold_fn: Predicate
    q_value_ratio_threshold: float = 1e9

    def __init__(self, threshold_fn: Predicate):
        self.threshold_fn = threshold_fn

    def has_valid_matches(self, chromatogram: 'TandemAnnotatedChromatogram', member_targets: Set[TargetType]) -> bool:
        has_matches = any([chromatogram.solutions_for(target, threshold_fn=self.threshold_fn)
                           for target in member_targets])
        return has_matches

    def validate_spectrum_match(self, spectrum_match: Union[SpectrumMatch, MultiScoreSpectrumMatch],
                                solution_set: SpectrumSolutionSet) -> bool:
        ## It would be desirable to be able to detect when the difference
        ## in explanations are of such radically different quality that
        ## any consideration of revision is a waste, but this can't be
        ## done reliably as a function of q-value. It might be feasible to do
        ## with score, but such a modification would be difficult to tune for
        ## different scoring algorithms simultaneously. Perhaps if we
        ## had a multidimensional threshold instead of a cascading threshold,
        ## but such a thresholding scheme would need to be estimated
        ## on the fly.
        # best_solution: MultiScoreSpectrumMatch = solution_set.best_solution()
        # if not self.threshold_fn(best_solution):
        #     return True
        # ratio = spectrum_match.q_value / best_solution.q_value
        # pass_threshold = ratio <= self.q_value_ratio_threshold
        # if not pass_threshold:
        #     self.log(
        #         f"... Rejecting revision of {spectrum_match.scan.id} from {best_solution.target} to {spectrum_match.target} due to ID confidence {spectrum_match.q_value}/{best_solution.q_value}")
        # return pass_threshold
        return True

    def can_rewrite_best_matches(self, chromatogram: 'TandemAnnotatedChromatogram', target: TargetType) -> bool:
        # The count of spectra we are able to actually interpret the best spectrum match of
        ssets_to_convert = 0
        # The count of spectra where we can reasonably revise the best match
        ssets_could_convert = 0
        for sset in chromatogram.tandem_solutions:
            if self.threshold_fn(sset.best_solution()):
                ssets_to_convert += 1
            try:
                match = sset.solution_for(target)
                if self.threshold_fn(match) and self.validate_spectrum_match(match, sset):
                    if not match.best_match:
                        ssets_could_convert += 1
            except KeyError:
                continue
        if ssets_to_convert == 0:
            return True
        return (ssets_could_convert / ssets_to_convert) >= 0.5

    def do_rewrite_best_matches(self, chromatogram: 'TandemAnnotatedChromatogram',
                                target: TargetType,
                                invalidated_targets: Set[TargetType]):
        for sset in chromatogram.tandem_solutions:
            try:
                match = sset.solution_for(target)
                if self.threshold_fn(match) and self.validate_spectrum_match(match, sset):
                    if not match.best_match:
                        sset.promote_to_best_match(match)
                else:
                    self.debug(
                        f"... Skipping invalidation of {sset.scan_id!r}, alternative {target} did not pass threshold.")
                    continue
            except KeyError:
                # TODO: Fill in missing match against the preferred target
                self.debug(
                    f"... Skipping invalidation of {sset.scan_id!r}, alternative {target} was not matched.")
                continue

            for invalid_target in invalidated_targets:
                try:
                    match = sset.solution_for(invalid_target)
                except KeyError:
                    continue
                # TODO: Is this really the right way to handle cases with totally
                # different peptide backbones? This should require a minimum of MS2 score/FDR
                # threshold passing
                if match.best_match:
                    self.debug(
                        f"... Revoking best match status of {match.target} for scan {match.scan_id!r}")
                match.best_match = False


class SignalUtilizationMS2RevisionValidator(MS2RevisionValidator):
    utilization_ratio_threshold: float = 0.6

    def __init__(self, threshold_fn: Predicate, utilization_ratio_threshold: float = 0.6):
        super().__init__(threshold_fn)
        self.utilization_ratio_threshold = utilization_ratio_threshold

    def validate_spectrum_match(self, spectrum_match: Union[SpectrumMatch, MultiScoreSpectrumMatch], solution_set: SpectrumSolutionSet) -> bool:
        baseline = super().validate_spectrum_match(spectrum_match, solution_set)
        if not baseline:
            return baseline
        best_solution: MultiScoreSpectrumMatch = solution_set.best_solution()
        if not self.threshold_fn(best_solution):
            return True
        ratio = spectrum_match.score_set.total_signal_utilization / \
            best_solution.score_set.total_signal_utilization
        pass_threshold = ratio >= self.utilization_ratio_threshold
        if baseline and not pass_threshold:
            self.log(
                f"... Rejecting revision of {spectrum_match.scan.id} from {best_solution.target} to {spectrum_match.target} due to signal ratio failing to pass threshold {ratio:0.3f}")
        return pass_threshold

    def has_valid_matches(self, chromatogram: 'TandemAnnotatedChromatogram', targets: Set[TargetType]) -> bool:
        has_passing_case = False
        # Loop over all targets and check if any of them manage to pass the ratio threshold,
        # preventing
        for target in targets:
            for sset in chromatogram.tandem_solutions:
                try:
                    solution: MultiScoreSpectrumMatch = sset.solution_for(
                        target)
                except KeyError:
                    continue

                # Don't bother checking solutions that don't pass FDR threshold predicate
                if not self.threshold_fn(solution):
                    continue

                if self.validate_spectrum_match(solution, sset):
                    # If a match passes the threshold, then something in this solution set is valid
                    has_passing_case = True
                    break
                else:
                    # Otherwise keep checking solution sets
                    continue

            # If we found a valid match, we're done.
            if has_passing_case:
                break

        return has_passing_case


class SpectrumMatchBackFiller(TaskBase):
    scorer: Type[SpectrumMatcherBase]
    match_args: Dict[str, Any]

    spectrum_match_cls: Type[SpectrumMatch]
    fdr_estimator: Optional[FDREstimatorBase]
    threshold_fn: Predicate
    mass_shifts: Optional[List[MassShiftBase]]

    scan_loader: ProcessedRandomAccessScanSource

    def __init__(self, scan_loader: ProcessedRandomAccessScanSource,
                 scorer: Type[SpectrumMatcherBase],
                 spectrum_match_cls: Type[SpectrumMatch],
                 id_maker,
                 threshold_fn: Predicate = lambda x: x.q_value < 0.05,
                 match_args=None,
                 fdr_estimator: Optional[FDREstimatorBase] = None,
                 mass_shifts: Optional[List[MassShiftBase]] = None):
        if match_args is None:
            match_args = {}
        self.scan_loader = scan_loader
        self.scorer = scorer
        self.match_args = match_args
        self.spectrum_match_cls = spectrum_match_cls
        self.id_maker = id_maker
        self.fdr_estimator = fdr_estimator
        self.threshold_fn = threshold_fn
        self.mass_shifts = mass_shifts

    def select_best_mass_shift_for(self, scan: ProcessedScan,
                                   structure: TargetType,
                                   mass_shifts: List[MassShiftBase]) -> Tuple[Optional[MassShiftBase], float]:
        observed_mass = scan.precursor_information.neutral_mass
        theoretical_mass = structure.total_mass
        delta = observed_mass - theoretical_mass

        best_shift = None
        best_shift_error = float('inf')

        for shift in mass_shifts:
            err = delta - shift.mass
            abs_err = abs(err)
            if abs_err < best_shift_error:
                best_shift = shift
                best_shift_error = abs_err
        return best_shift, best_shift_error

    def id_for_structure(self, structure: TargetType, reference: TargetType):
        structure.protein_relation = reference.protein_relation
        structure.id = self.id_maker(structure, reference)
        return structure

    def evaluate_spectrum(self, scan_id: str, structure: TargetType, mass_shifts: List[MassShiftBase]) -> SpectrumMatcherBase:
        scan = self.scan_loader.get_scan_by_id(scan_id)
        best_shift, best_shift_error = self.select_best_mass_shift_for(
            scan, structure, mass_shifts)
        match = self.scorer.evaluate(
            scan, structure, mass_shift=best_shift, **self.match_args)
        return match

    def add_matches_to_solution_set(self, sset: SpectrumSolutionSet,
                                    targets: List[TargetType],
                                    mass_shifts: Optional[List[MassShiftBase]] = None,
                                    should_promote: bool=False) -> List[Tuple[SpectrumSolutionSet, SpectrumMatch, bool]]:
        if mass_shifts is None:
            mass_shifts = self.mass_shifts

        solution_set_match_pairs = []

        scan_id = sset.scan_id
        for inst in targets:
            matched = self.evaluate_spectrum(
                scan_id, inst, mass_shifts)
            match = self.spectrum_match_cls.from_match_solution(
                matched)
            match.scan = sset.scan
            sset.insert(0, match)
            solution_set_match_pairs.append((sset, match, False))

        sset.sort()
        sset.mark_top_solutions()
        if self.fdr_estimator is not None:
            self.fdr_estimator.score_all(sset)

        for inst in targets:
            match = sset.solution_for(inst)
            if self.threshold_fn(match) and should_promote:
                match.best_match = True
                sset.promote_to_best_match(match)
        return solution_set_match_pairs


class SpectrumMatchInvalidatinCallback:
    target_to_find: TargetType
    threshold_fn: Callable[[SpectrumMatch], bool]

    def __init__(self, target_to_find: TargetType, threshold_fn: Callable[[SpectrumMatch], bool]):
        self.target_to_find = target_to_find
        self.threshold_fn = threshold_fn

    def __call__(self, sset: SpectrumSolutionSet, sm: SpectrumMatch) -> bool:
        try:
            preferred_sm = sset.solution_for(self.target_to_find)
            if not self.threshold_fn(preferred_sm):
                return True
            return False
        except KeyError:
            return True


class MatchMarkResult(NamedTuple):
    targets: Set[TargetType]
    match_count: int

    def __add__(self, other: 'MatchMarkResult'):
        if other is None:
            return self
        return self.__class__(
            self.targets | other.targets,
            self.match_count + other.match_count
        )

    @classmethod
    def empty(cls):
        return cls(set(), 0)


InvalidationFilter = Callable[[SpectrumSolutionSet, SpectrumMatch], bool]


class RevisionSummary(NamedTuple):
    accepted: bool
    rejected_matches: Optional[MatchMarkResult] = None
    invalidated_matches: Optional[MatchMarkResult] = None

    @property
    def rejected_match_count(self):
        if not self.rejected_matches:
            return 0
        return self.rejected_matches.match_count

    @property
    def invalidated_match_count(self):
        if not self.invalidated_matches:
            return 0
        return self.invalidated_matches.match_count


class SpectrumMatchUpdater(SpectrumMatchBackFiller):
    """Generate updates to :class:`TandemAnnotatedChromatogram` from chromatogram reviser like an RT predictor.

    This type needs to:
        1. Generate new spectrum matches using the same scorer if a solution set lacks
           a match to the suggested target.
        2. Re-calculate the FDR of the new matches if needed.
        3. Generate a new unique ID for each source peptide that a glycopeptide matching
           the revised glycoform could come from, that is consistent with the previous
           ID space that other glycopeptides used.
        4. Update the assigned entity of each assigned chromatogram so that the correct
           structure is the "best match", discarding the :class:`SolutionEntry` for the
           incorrect assignment *without* deleting the spectrum matches to that themselves.
    """

    spectrum_match_cls: Type[SpectrumMatch]
    threshold_fn: Callable[[SpectrumMatch], bool]
    fdr_estimator: FDREstimatorBase
    match_args: Dict[str, Any]

    retention_time_delta: float
    retention_time_model: 'RetentionTimeModelEnsemble'

    def __init__(self, scan_loader,
                 scorer,
                 spectrum_match_cls,
                 id_maker,
                 threshold_fn=lambda x: x.q_value < 0.05,
                 match_args=None,
                 fdr_estimator=None, retention_time_model=None,
                 retention_time_delta=0.35):
        if match_args is None:
            match_args = {}

        super().__init__(
            scan_loader,
            scorer,
            spectrum_match_cls,
            id_maker,
            threshold_fn,
            match_args,
            fdr_estimator)

        self.retention_time_model = retention_time_model
        self.retention_time_delta = retention_time_delta

    def _find_identical_keys_and_matches(self, chromatogram: 'SpectrumMatchSolutionCollectionBase',
                                         structure: TargetType) -> Tuple[Set[TargetType], List[Tuple[SpectrumSolutionSet, SpectrumMatch]]]:
        structures: Set[TargetType] = set()
        key = str(structure)
        spectrum_matches: List[Tuple[SpectrumSolutionSet, SpectrumMatch]] = []
        for sset in chromatogram.tandem_solutions:
            for sm in sset:
                if str(sm.target) == key:
                    structures.add(sm.target)
                    spectrum_matches.append((sset, sm))
        return structures, spectrum_matches

    def find_identical_peptides(self, chromatogram: 'SpectrumMatchSolutionCollectionBase',
                                structure: TargetType) -> Set[TargetType]:
        structures, _spectrum_matches = self._find_identical_keys_and_matches(
            chromatogram, structure)
        return structures

    def find_identical_peptides_and_mark(self, chromatogram: 'SpectrumMatchSolutionCollectionBase',
                                         structure: TargetType,
                                         best_match: bool = False,
                                         valid: bool = False,
                                         reason: Optional[str] = None,
                                         filter_fn: Optional[InvalidationFilter] = None
                                         ) -> MatchMarkResult:

        structures, spectrum_matches = self._find_identical_keys_and_matches(
            chromatogram, structure)
        k = 0
        for sset, sm in spectrum_matches:
            if filter_fn and filter_fn(sset, sm):
                self.debug(
                    f"... NOT Marking {sm.scan_id} -> {sm.target} Valid = {valid},"
                    f" Best Match = {best_match}; Reason: {reason}")
                continue
            self.debug(
                f"... Marking {sm.scan_id} -> {sm.target} Valid = {valid}, Best Match = {best_match}; Reason: {reason}")
            sm.valid = valid
            sm.best_match = best_match
            # Do not want to invalidate order as that would force re-sorting
            # and that would break any "promoted" solution whose score isn't
            # the top one.
            sset._invalidate(invalidate_order=False)
            k += 1
        return MatchMarkResult(structures, k)

    def make_target_filter(self, target_to_find: TargetType) -> InvalidationFilter:
        filter_fn = SpectrumMatchInvalidatinCallback(
            target_to_find, self.threshold_fn)

        return filter_fn

    def get_spectrum_solution_sets(self, revision: Union['TandemAnnotatedChromatogram',
                                                         'TandemSolutionsWithoutChromatogram',
                                                         'ChromatogramProxy'],
                                   chromatogram: Optional['SpectrumMatchSolutionCollectionBase'] = None,
                                   invalidate_reference: bool = True):
        if chromatogram is None:
            chromatogram = revision.source

        reference = chromatogram.structure
        mass_shifts = revision.mass_shifts
        revised_structure = revision.structure

        if invalidate_reference:
            reference_peptides, _count = self.find_identical_peptides_and_mark(
                chromatogram,
                reference,
                reason=f"superceded by {revision.structure}")
        else:
            reference_peptides = self.find_identical_peptides(
                chromatogram, reference)

        revised_solution_sets = self.collect_matches(
            chromatogram,
            revised_structure,
            mass_shifts,
            reference_peptides,
            should_promote=invalidate_reference)

        if self.fdr_estimator is not None:
            for sset in chromatogram.tandem_solutions:
                self.fdr_estimator.score_all(sset)
        return revised_solution_sets

    def collect_candidate_solutions_for_rt_validation(self, chromatogram: Union['TandemAnnotatedChromatogram',
                                                                                'TandemSolutionsWithoutChromatogram'],
                                                      structure: TargetType,
                                                      rt_score: float,
                                                      overridden: Optional[TargetType] = None) -> Tuple[
                                                        List[SolutionEntry],
                                                        List[SolutionEntry],
                                                        List[SolutionEntry],
                                                        Set[TargetType]
                                                    ]:
        from glycresoft.scoring.elution_time_grouping import GlycopeptideChromatogramProxy
        from .chromatogram import TandemSolutionsWithoutChromatogram

        representers: List[SolutionEntry] = chromatogram.compute_representative_weights(
            self.threshold_fn)

        keep: List[SolutionEntry] = []
        match: List[SolutionEntry] = []
        rejected: List[SolutionEntry] = []
        invalidated: Set[TargetType] = set()
        alternative_rt_scores: Dict[TargetType, float] = {}

        for r in representers:
            if str(r.solution) == overridden:
                rejected.append(r)
            elif str(r.solution) == structure:
                match.append(r)
            else:
                if self.retention_time_model and not isinstance(chromatogram, TandemSolutionsWithoutChromatogram):
                    if r.solution not in alternative_rt_scores:
                        proxy = GlycopeptideChromatogramProxy.from_chromatogram(
                            chromatogram)
                        proxy.structure = r.solution
                        alt_rt_score = alternative_rt_scores[r.solution] = self.retention_time_model.score_interval(
                            proxy, 0.01)
                    else:
                        alt_rt_score = alternative_rt_scores[r.solution]
                    if np.isnan(alt_rt_score) and not np.isnan(rt_score):
                        self.log(
                            f"RT score for alternative {r.solution} is NaN, reference {structure}"
                            f" RT score is {rt_score:0.3f}")
                    elif not np.isnan(alt_rt_score) and np.isnan(rt_score):
                        self.log(
                            f"RT score alternative {r.solution} is {alt_rt_score:0.3f}, reference"
                            f" {structure} RT score is NaN")
                    # If either is NaN, this is always false, so we never invalidate a NaN-predicted case.
                    if (rt_score - alt_rt_score) > self.retention_time_delta:
                        invalidated.add(r.solution)
                        continue
                keep.append(r)

        return match, keep, rejected, invalidated

    def process_revision(self, revision: 'GlycopeptideChromatogramProxy',
                         chromatogram: Union[
                            'TandemAnnotatedChromatogram',
                            'TandemSolutionsWithoutChromatogram'] = None) -> 'TandemAnnotatedChromatogram':

        if chromatogram is None:
            chromatogram = revision.source

        reference = chromatogram.structure
        revised_structure = revision.structure
        assert reference != revised_structure

        _revised_solution_sets = self.get_spectrum_solution_sets(
            revision,
            chromatogram,
            invalidate_reference=True
        )

        representers: List[SolutionEntry] = chromatogram.compute_representative_weights(
            self.threshold_fn)

        revised_rt_score = None
        if self.retention_time_model:
            revised_rt_score = self.retention_time_model.score_interval(
                revision, 0.01)

        match, keep, rejected, invalidated = self.collect_candidate_solutions_for_rt_validation(
            chromatogram, revised_structure, revised_rt_score, reference)

        if not match:
            self.log("... Failed to identify alternative %s for %s passing the threshold" % (
                revised_structure, reference))
            return chromatogram, RevisionSummary(False, None, None)

        alternative_filter_fn = self.make_target_filter(revised_structure)

        rejected_track = MatchMarkResult(set(), 0)
        for reject in rejected:
            marks = self.find_identical_peptides_and_mark(
                chromatogram,
                reject.solution,
                reason=f"Rejecting {reject} in favor of {revised_structure}",
                filter_fn=alternative_filter_fn)
            rejected_track = rejected_track + marks

        invalidated_track = MatchMarkResult(set(), 0)
        for invalid in invalidated:
            marks = self.find_identical_peptides_and_mark(
                chromatogram,
                invalid,
                reason=f"... Invalidating {invalid} in favor of {revised_structure}",
                filter_fn=alternative_filter_fn)
            invalidated_track = invalidated_track + marks

        marks = self.find_identical_peptides_and_mark(
            chromatogram,
            reference,
            reason=f"superceded by {revision.structure}")
        rejected_track = rejected_track + marks

        representers = chromatogram.filter_entries(match + keep)
        # When the input is a ChromatogramSolution, we have to explicitly
        # set the attribute on the wrapped chromatogram object, not the wrapper
        # or else it will be lost during later unwrapping.
        if isinstance(chromatogram, ChromatogramSolution):
            chromatogram.chromatogram.representative_solutions = representers
        else:
            chromatogram.representative_solutions = representers

        # This mutation is safe to call directly since it passes through the
        # universal __getattr__ wrapper
        chromatogram.assign_entity(match[0])
        return chromatogram, RevisionSummary(True, rejected_track, invalidated_track)

    def collect_matches(self,
                        chromatogram: 'TandemAnnotatedChromatogram',
                        structure: TargetType,
                        mass_shifts: List[MassShiftBase],
                        reference_peptides: List[TargetType],
                        should_promote: bool=False) -> List[Tuple[SpectrumSolutionSet, SpectrumMatch, bool]]:
        """Collect spectrum matches to the new `structure` target, and all identical solutions from different sources for all spectra in `chromatogram`.

        Parameters
        ----------
        chromatogram : :class:`~.TandemAnnotatedChromatogram`
            The solution collection being traversed.
        structure : :class:`~.TargetType`
            The new structure we are going to assign to `chromatogram` and wish to
            ensure there are spectrum matches for.
        mass_shifts : :class:`list` of :class:`~.MassShiftBase`
            The mass shifts which are legal to consider for matching `structure` to spectra
            with.
        reference_peptides : :class:`set` of :class:`~.TargetType`
            Other sources of the peptide for `structure` to ensure that all are reported together.
            Relies on :attr:`id_maker` to re-create appropriately keyed glycoforms.
        should_promote : :class:`bool`
            When :const:`True`, matches to ``structure`` are promoted to best-match status if they
            pass the :attr:`threshold_fn`.
        """
        solution_set_match_pairs = []
        if structure.id is None or reference_peptides:
            instances = self.id_maker(structure, reference_peptides)
        else:
            instances = []
        for sset in chromatogram.tandem_solutions:
            try:
                match = sset.solution_for(structure)
                if not match.best_match and self.threshold_fn(match) and should_promote:
                    sset.promote_to_best_match(match)
                solution_set_match_pairs.append((sset, match, True))
            except KeyError:
                solution_set_match_pairs.extend(
                    self.add_matches_to_solution_set(
                        sset,
                        instances,
                        mass_shifts,
                        should_promote=should_promote
                    )
                )
        return solution_set_match_pairs

    def __call__(self, revision, chromatogram=None):
        return self.process_revision(revision, chromatogram=chromatogram)

    def ensure_q_values(self, chromatogram):
        for sset in chromatogram.tandem_solutions:
            for sm in sset:
                if sm.q_value is None and self.fdr_estimator is not None:
                    self.debug("... Computing q-value for %s+%s @ %r" %
                               (sm.target, sm.mass_shift.name, sm.scan_id))
                    self.fdr_estimator.score(sm)

    def affirm_solution(self, chromatogram: 'TandemAnnotatedChromatogram') -> 'TandemAnnotatedChromatogram':
        reference_peptides = self.find_identical_peptides(
            chromatogram, chromatogram.entity)

        _matches = self.collect_matches(
            chromatogram,
            chromatogram.entity,
            chromatogram.mass_shifts,
            reference_peptides,
            should_promote=True
        )
        revised_rt_score = None
        if self.retention_time_model:
            revised_rt_score = self.retention_time_model.score_interval(
                chromatogram, 0.01)

        self.ensure_q_values(chromatogram)
        match, _keep, rejected, invalidated = self.collect_candidate_solutions_for_rt_validation(
            chromatogram, chromatogram.entity, revised_rt_score, None)

        filter_fn = self.make_target_filter(chromatogram.entity)

        rejected_track = MatchMarkResult.empty()
        for reject in rejected:
            marks = self.find_identical_peptides_and_mark(
                chromatogram,
                reject.solution,
                reason=f"Rejecting {reject} in favor of {chromatogram.entity}",
                filter_fn=filter_fn
            )
            rejected_track = rejected_track + marks

        invalidated_track = MatchMarkResult.empty()
        for invalid in invalidated:
            marks = self.find_identical_peptides_and_mark(
                chromatogram,
                invalid,
                reason=f"Invalidating {invalid} in favor of {chromatogram.entity}",
                filter_fn=filter_fn
            )
            invalidated_track = invalidated_track + marks

        if not match:
            self.log("... Failed to affirm %s @ %0.2f passing the threshold" % (
                chromatogram.entity, getattr(chromatogram, "apex_time", -1)))
            affirmed = False
        else:
            affirmed = True
        return chromatogram, RevisionSummary(affirmed, rejected_track, invalidated_track)



# cython: embedsignature=True
from collections import defaultdict

cimport cython
from cpython cimport PyObject
from cpython cimport PyList_Append, PyList_Size, PyList_GetItem
from cpython cimport PySet_Add
from cpython cimport PyDict_GetItem, PyDict_SetItem, PyDict_Contains
from cpython cimport PyInt_AsLong

import numpy as np
cimport numpy as np

from ms_deisotope._c.peak_set cimport DeconvolutedPeak

from glycopeptidepy._c.structure.fragment cimport FragmentBase, SimpleFragment, PeptideFragment


np.import_array()

@cython.final
@cython.freelist(1000000)
cdef class PeakFragmentPair(object):

    @staticmethod
    cdef PeakFragmentPair _create(DeconvolutedPeak peak, fragment):
        cdef PeakFragmentPair self = PeakFragmentPair.__new__(PeakFragmentPair)
        self.peak = peak
        self.fragment = fragment
        self.fragment_name = fragment.name
        self._hash = hash(peak.mz)
        return self

    @staticmethod
    cdef PeakFragmentPair _create_simple(DeconvolutedPeak peak, FragmentBase fragment):
        cdef PeakFragmentPair self = PeakFragmentPair.__new__(PeakFragmentPair)
        self.peak = peak
        self.fragment = fragment
        self.fragment_name = fragment.get_name()
        self._hash = hash(peak.mz)
        return self

    def __init__(self, peak, fragment):
        self.peak = peak
        self.fragment = fragment
        self.fragment_name = fragment.name
        self._hash = hash(self.peak.mz)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(self, PeakFragmentPair):
            return False
        return (self.peak == (<PeakFragmentPair>other).peak) and (
            self.fragment_name == (<PeakFragmentPair>other).fragment_name)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return self._hash

    def __reduce__(self):
        return self.__class__, (self.peak, self.fragment)

    def clone(self):
        return self.__class__(self.peak, self.fragment)

    def __repr__(self):
        return "PeakFragmentPair(%r, %r)" % (self.peak, self.fragment)

    def __iter__(self):
        yield self.peak
        yield self.fragment

    @cython.cdivision(True)
    cpdef double mass_accuracy(self):
        return (self.peak.neutral_mass - self.fragment.mass) / self.fragment.mass


@cython.final
@cython.freelist(1000000)
cdef class PeakPairTransition(object):

    @staticmethod
    cdef PeakPairTransition _create(DeconvolutedPeak start, DeconvolutedPeak end, object annotation):
        cdef PeakPairTransition self = PeakPairTransition.__new__(PeakPairTransition)
        self.start = start
        self.end = end
        self.annotation = annotation
        # The indices of the start peak and end peak
        self.key = (self.start.index.neutral_mass, self.end.index.neutral_mass)
        self._hash = hash(self.key)
        return self

    def __init__(self, start, end, annotation):
        self.start = start
        self.end = end
        self.annotation = annotation
        # The indices of the start peak and end peak
        self.key = (self.start.index.neutral_mass, self.end.index.neutral_mass)
        self._hash = hash(self.key)

    def __reduce__(self):
        return self.__class__, (self.start, self.end, self.annotation)

    def __eq__(self, other):
        cdef:
            PeakPairTransition typed_other
        if other is None:
            return False
        else:
            typed_other = <PeakPairTransition?>other
        # if self.key != typed_other.key:
        #     return False
        if not self.start._eq(typed_other.start):
            return False
        elif not self.end._eq(typed_other.end):
            return False
        elif self.annotation != typed_other.annotation:
            return False
        return True

    def __iter__(self):
        yield self.start
        yield self.end
        yield self.annotation

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return ("{self.__class__.__name__}({self.start.neutral_mass:0.3f} "
                "-{self.annotation}-> {self.end.neutral_mass:0.3f})").format(self=self)


cdef class SpectrumGraph(object):
    def __init__(self):
        self.transitions = set()
        self.by_first = defaultdict(list)
        self.by_second = defaultdict(list)

    cpdef add(self, DeconvolutedPeak p1, DeconvolutedPeak p2, object annotation):
        if p1._index.neutral_mass > p2._index.neutral_mass:
            temp = p2
            p2 = p1
            p1 = temp
        trans = PeakPairTransition._create(p1, p2, annotation)
        PySet_Add(self.transitions, trans)
        self.by_first[trans.key[0]].append(trans)
        self.by_second[trans.key[1]].append(trans)

    def __iter__(self):
        return iter(self.transitions)

    def __len__(self):
        return len(self.transitions)

    def __repr__(self):
        return "{self.__class__.__name__}({size})".format(
            self=self, size=len(self.transitions))

    def _get_maximum_index(self):
        try:
            trans = max(self.transitions, key=lambda x: x.key[1])
            return trans.key[1]
        except ValueError:
            return 0

    def adjacency_matrix(self):
        n = self._get_maximum_index() + 1
        A = np.zeros((n, n))
        for trans in self.transitions:
            A[trans.key] = 1
        return A

    cpdef list topological_sort(self, adjacency_matrix=None):
        if adjacency_matrix is None:
            adjacency_matrix = self.adjacency_matrix()
        else:
            adjacency_matrix = adjacency_matrix.copy()

        waiting = set()
        for i in range(adjacency_matrix.shape[0]):
            # Check for incoming edges. If no incoming
            # edges, add to the waiting set
            if adjacency_matrix[:, i].sum() == 0:
                waiting.add(i)
        ordered = list()
        while waiting:
            ix = waiting.pop()
            ordered.append(ix)
            outgoing_edges = adjacency_matrix[ix, :]
            indices_of_neighbors = np.nonzero(outgoing_edges)[0]
            # For each outgoing edge
            for neighbor_ix in indices_of_neighbors:
                # Remove the edge
                adjacency_matrix[ix, neighbor_ix] = 0
                # Test for incoming edges
                if (adjacency_matrix[:, neighbor_ix] == 0).all():
                    waiting.add(neighbor_ix)
        if adjacency_matrix.sum() > 0:
            raise ValueError("%d edges left over" % (adjacency_matrix.sum(),))
        else:
            return ordered

    def path_lengths(self):
        adjacency_matrix = self.adjacency_matrix()
        distances = np.zeros(self._get_maximum_index() + 1)
        for ix in self.topological_sort():
            incoming_edges = np.nonzero(adjacency_matrix[:, ix])[0]
            if incoming_edges.shape[0] == 0:
                distances[ix] = 0
            else:
                longest_path_so_far = distances[incoming_edges].max()
                distances[ix] = longest_path_so_far + 1
        return distances

    cpdef list paths_starting_at(self, ix, int limit=-1):
        paths = []
        for trans in self.by_first[ix]:
            paths.append([trans])
        finished_paths = []
        while True:
            extended_paths = []
            for path in paths:
                terminal = path[-1]
                edges = self.by_first[terminal.key[1]]
                if not edges:
                    finished_paths.append(path)
                for trans in edges:
                    extended_paths.append(path + [trans])
            paths = extended_paths
            if len(paths) == 0:
                break
        return self.transitive_closure(finished_paths, limit)

    cpdef list paths_ending_at(self, ix, int limit=-1):
        paths = []
        for trans in self.by_second[ix]:
            paths.append([trans])
        finished_paths = []
        while True:
            extended_paths = []
            for path in paths:
                terminal = path[0]
                edges = self.by_second[terminal.key[0]]
                if not edges:
                    finished_paths.append(path)
                for trans in edges:
                    new_path = [trans]
                    new_path.extend(path)
                    extended_paths.append(new_path)
            paths = extended_paths
            if len(paths) == 0:
                break
        return self.transitive_closure(finished_paths, limit)

    cpdef list transitive_closure(self, list paths, int limit=-1):
        cdef:
            list path, index_list
            list node_sets, keep, node_sets_items
            set node_set, other_node_set
            dict by_node
            PyObject* tmp
            object conv
            PeakPairTransition node
            size_t i, j, n, m, k, q, v
            bint is_enclosed
        paths = sorted(paths, key=len, reverse=True)
        # precompute node_sets for each path
        node_sets = []
        by_node = {}
        for i in range(PyList_Size(paths)):
            path = <list>PyList_GetItem(paths, i)
            # track all nodes by index in this path in node_set
            node_set = set()
            for j in range(PyList_Size(path)):
                node = <PeakPairTransition>PyList_GetItem(path, j)
                # add the start node index and end node index to the
                # set of nodes on this path
                # node_set.update(node.key)
                if j == 0:
                    conv = node.start._index.neutral_mass
                    node_set.add(conv)
                    tmp = PyDict_GetItem(by_node, conv)
                    if tmp == NULL:
                        index_list = []
                        PyDict_SetItem(by_node, conv, index_list)
                    else:
                        index_list = <list>tmp
                    index_list.append(i)
                conv = node.end._index.neutral_mass
                node_set.add(conv)
                tmp = PyDict_GetItem(by_node, conv)
                if tmp == NULL:
                    index_list = []
                    PyDict_SetItem(by_node, conv, index_list)
                else:
                    index_list = <list>tmp
                index_list.append(i)
            node_sets.append(node_set)

        keep = []
        k = 0
        for i in range(PyList_Size(paths)):
            path = <list>PyList_GetItem(paths, i)
            is_enclosed = False
            node_set = <set>PyList_GetItem(node_sets, i)
            n = len(node_set)
            node = <PeakPairTransition>PyList_GetItem(path, 0)
            conv = node.start._index.neutral_mass
            index_list = <list>PyDict_GetItem(by_node, conv)
            for q in range(PyList_Size(index_list)):
                j = PyInt_AsLong(<object>PyList_GetItem(index_list, q))
                if i == j:
                    continue
                other_node_set = <set>PyList_GetItem(node_sets, j)
                m = len(other_node_set)
                if m < n:
                    break
                if node_set < other_node_set:
                    is_enclosed = True
                    break
            if not is_enclosed:
                keep.append(path)
                k += 1
            if limit > 0 and limit == k:
                break
        return keep

    def longest_paths(self, int limit=-1):
        cdef:
            list paths, segment
        # get all distinct paths
        paths = []
        for ix in np.argsort(self.path_lengths())[::-1]:
            segment = self.paths_ending_at(ix, limit)
            paths.extend(segment)
            if PyList_Size(segment) == 0:
                break
            # remove redundant paths
            paths = self.transitive_closure(paths, limit)
            if limit > 0:
                if PyList_Size(paths) > limit:
                    break
        return paths


@cython.freelist(10000)
cdef class _FragmentIndexBase(object):
    def __init__(self, fragment_set):
        self.fragment_set = fragment_set
        self._mapping = None

    cdef int _create_mapping(self):
        return 0

    @property
    def mapping(self):
        if self._mapping is None:
            self._create_mapping()
        return self._mapping

    cdef void invalidate(self):
        self._mapping = None

    cdef list getitem(self, object key):
        cdef:
            PyObject* ptemp
        if self._mapping is None:
            self._create_mapping()
        ptemp = PyDict_GetItem(self._mapping, key)
        if ptemp == NULL:
            return []
        else:
            return <list>ptemp

    cdef bint has_key(self, object key):
        if self._mapping is None:
            self._create_mapping()
        return PyDict_Contains(self._mapping, key)

    def __getitem__(self, key):
        return self.getitem(key)

    def __iter__(self):
        return iter(self.mapping)

    def items(self):
        return self.mapping.items()

    def keys(self):
        return self.mapping.keys()

    def values(self):
        return self.mapping.values()

    def __len__(self):
        return len(self.mapping)

    def __contains__(self, key):
        return self.has_key(key)

    def __str__(self):
        return str(self.mapping)


cdef class ByFragmentIndex(_FragmentIndexBase):

    @staticmethod
    cdef ByFragmentIndex _create(FragmentMatchMap fragment_set):
        cdef ByFragmentIndex self = ByFragmentIndex.__new__(ByFragmentIndex)
        self.fragment_set = fragment_set
        self._mapping = None
        return self

    cdef int _create_mapping(self):
        cdef:
            object obj
            PyObject* ptemp
            PeakFragmentPair pair
            list bucket
        self._mapping = {}
        for obj in self.fragment_set.members:
            pair = <PeakFragmentPair>obj
            ptemp = PyDict_GetItem(self._mapping, pair.fragment)
            if ptemp == NULL:
                bucket = [pair.peak]
                PyDict_SetItem(self._mapping, pair.fragment, bucket)
            else:
                bucket = <list>ptemp
                bucket.append(pair.peak)
        return 0


cdef class ByPeakIndex(_FragmentIndexBase):

    @staticmethod
    cdef ByPeakIndex _create(FragmentMatchMap fragment_set):
        cdef ByPeakIndex self = ByPeakIndex.__new__(ByPeakIndex)
        self.fragment_set = fragment_set
        self._mapping = None
        return self

    cdef int _create_mapping(self):
        cdef:
            object obj
            PyObject* ptemp
            PeakFragmentPair pair
            list bucket
        self._mapping = {}
        for obj in self.fragment_set.members:
            pair = <PeakFragmentPair>obj
            ptemp = PyDict_GetItem(self._mapping, pair.peak)
            if ptemp == NULL:
                bucket = [pair.fragment]
                PyDict_SetItem(self._mapping, pair.peak, bucket)
            else:
                bucket = <list>ptemp
                bucket.append(pair.fragment)
        return 0


cdef class ByPeakIndexIndex(_FragmentIndexBase):

    @staticmethod
    cdef ByPeakIndexIndex _create(FragmentMatchMap fragment_set):
        cdef ByPeakIndexIndex self = ByPeakIndexIndex.__new__(ByPeakIndexIndex)
        self.fragment_set = fragment_set
        self._mapping = None
        return self

    cdef int _create_mapping(self):
        cdef:
            object obj
            PyObject* ptemp
            PeakFragmentPair pair
            DeconvolutedPeak peak
            list bucket
            dict mapping
        mapping = self._mapping = {}
        for obj in self.fragment_set.members:
            pair = <PeakFragmentPair>obj
            peak = pair.peak
            ptemp = PyDict_GetItem(mapping, peak._index.neutral_mass)
            if ptemp == NULL:
                bucket = [pair.fragment]
                PyDict_SetItem(mapping, peak._index.neutral_mass, bucket)
            else:
                bucket = <list>ptemp
                PyList_Append(bucket, pair.fragment)
        return 0



@cython.final
@cython.freelist(1000000)
cdef class FragmentMatchMap(object):
    def __init__(self):
        self.members = set()
        self.by_fragment = ByFragmentIndex._create(self)
        self.by_peak = ByPeakIndex._create(self)
        self.by_peak_index = ByPeakIndexIndex._create(self)

    cpdef add(self, peak, fragment=None):
        if fragment is not None:
            if isinstance(fragment, FragmentBase):
                peak = PeakFragmentPair._create_simple(<DeconvolutedPeak>peak, <FragmentBase>fragment)
            else:
                peak = PeakFragmentPair._create(<DeconvolutedPeak>peak, fragment)
        PySet_Add(self.members, peak)

    cdef void _add_direct(self, PeakFragmentPair pair):
        PySet_Add(self.members, pair)

    def pairs_by_name(self, name):
        pairs = []
        for pair in self.members:
            if pair.fragment_name == name:
                pairs.append(pair)
        return pairs

    cpdef list fragments_for(self, peak):
        return self.by_peak.getitem(peak)

    cpdef list peaks_for(self, fragment):
        return self.by_fragment.getitem(fragment)

    def __eq__(self, other):
        return self.members == other.members

    def __ne__(self, other):
        return self.members != other.members

    def __iter__(self):
        return iter(self.members)

    def __len__(self):
        return len(self.members)

    def items(self):
        for peak, fragment in self.members:
            yield fragment, peak

    def values(self):
        for pair in self.members:
            yield pair.peak

    cpdef set fragments(self):
        cdef:
            set frags
            object obj
            PeakFragmentPair pair
        frags = set()
        for obj in self.members:
            pair = <PeakFragmentPair>obj
            frags.add(pair.fragment)
        return frags

    def remove_fragment(self, fragment):
        peaks = self.peaks_for(fragment)
        for peak in peaks:
            self.members.remove(PeakFragmentPair(peak, fragment))
        self.by_fragment.invalidate()
        self.by_peak.invalidate()

    def remove_peak(self, peak):
        fragments = self.fragments_for(peak)
        for fragment in fragments:
            self.members.remove(PeakFragmentPair(peak, fragment))
        self.by_peak.invalidate()
        self.by_fragment.invalidate()

    cpdef FragmentMatchMap copy(self):
        cdef FragmentMatchMap inst = self.__class__()
        for case in self.members:
            inst.add(case)
        return inst

    cpdef FragmentMatchMap clone(self):
        return self.copy()

    def __repr__(self):
        return "FragmentMatchMap(%s)" % (', '.join(
            f.name for f in self.fragments()),)

    cpdef clear(self):
        self.members.clear()
        self.invalidate()

    cpdef invalidate(self):
        self.by_fragment.invalidate()
        self.by_peak.invalidate()

    cpdef set peak_indices(self):
        cdef:
            PeakFragmentPair pfp
            DeconvolutedPeak peak
            set result

        result = set()
        for obj in self.members:
            pfp = <PeakFragmentPair>obj
            peak = pfp.peak
            result.add(peak._index.neutral_mass)
        return result